/* BayesX - Software for Bayesian Inference in
Structured Additive Regression Models.
Copyright (C) 2011  Christiane Belitz, Andreas Brezger,
Thomas Kneib, Stefan Lang, Nikolaus Umlauf

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA. */

#include "distr_gamlss_patrick.h"
#include "gsl/gsl_randist.h"
#include "gsl/gsl_cdf.h"


namespace MCMC
{

    //------------------------------------------------------------------------------
    //-------------- CLASS: DISTR_sndp_alpha ---------------------------------------
    //------------------------------------------------------------------------------

    // Constructor:
    DISTR_sndp_alpha::DISTR_sndp_alpha(GENERAL_OPTIONS * o,
                                       const datamatrix & r,
                                       const datamatrix & w)
    : DISTR_gamlss(o,r,2,w) // <-----
    {
        family = "Skew Normal Distribution, Direct Parametrization - alpha";
        outpredictor = true;
        outexpectation = false; // only true if main parameter
        predictor_name = "alpha";

        // linpredminlimit=-10; // necessary for log-link
        // linpredmaxlimit=15; // necessary for log-link

        // updateIWLS = false; // only false and necessary if gibbs sampling?
    }


    // Copy Constructor:
    DISTR_sndp_alpha::DISTR_sndp_alpha(const DISTR_sndp_alpha & nd)
    : DISTR_gamlss(DISTR_gamlss(nd))
    {

    }

    // Operator:
    const DISTR_sndp_alpha & DISTR_sndp_alpha::operator=(
                                                         const DISTR_sndp_alpha & nd)
    {
        if (this==&nd)
            return *this;
        DISTR_gamlss::operator=(DISTR_gamlss(nd));
        return *this;
    }



    double DISTR_sndp_alpha::get_intercept_start(void)
    {
        return 0;
    }

    void DISTR_sndp_alpha::compute_param_mult(vector<double *>  linpred,double * param)
    {
        *param = (*linpred[0]); // <--- c.f. linpred info
    }



    double DISTR_sndp_alpha::loglikelihood_weightsone(double * response,
                                                      double * linpred)
    {
        if (counter==0)
        {
            set_worklin();
            // if((*linpred) > 10.0){
            //   (*linpred) = 10.0;
            // }
            // else if((*linpred) < - 10.0){
            //   (*linpred) = -10.0;
            // }
        }

        double x = (*response);
        double xi = *worklin[1];
        double om = exp(*worklin[0]);
        double al = (*linpred);

        double l;

        l = log(randnumbers::Phi2(al*(x-xi)/om)); // calculate log-likelihood (possibly simplified)

        modify_worklin();

        return l;

    }


    void DISTR_sndp_alpha::compute_iwls_wweightschange_weightsone(
                                                                  double * response,
                                                                  double * linpred,
                                                                  double * workingweight,
                                                                  double * workingresponse,
                                                                  double & like,
                                                                  const bool & compute_like)
    {
        if (counter==0)
        {
            set_worklin();
            // if((*linpred) > 10.0){
            //   (*linpred) = 10.0;
            // }
            // else if((*linpred) < - 10.0){
            //   (*linpred) = -10.0;
            // }
        }

        double x = (*response);
        double xi = (*worklin[1]);
        double om = exp(*worklin[0]);
        double al = (*linpred);

        double z = (x - xi)/om;

        double nu = randnumbers::phi(al*z) / randnumbers::Phi2(al*z) * z; // v

        *workingweight = supportfunctions::sn_a2(al); // w

        *workingresponse = *linpred + nu/(*workingweight);

        if (compute_like)
        {

            like += log(randnumbers::Phi2(al*z)); // loglikelihood

        }


        modify_worklin();

    }


    void DISTR_sndp_alpha::outoptions(void)
    {
        DISTR::outoptions();
        optionsp->out("Link function (alpha): identity\n");
        optionsp->out("\n");
        optionsp->out("\n");
    }


    void DISTR_sndp_alpha::update_end(void)
    {


        // helpmat1 stores (eta_mu)

        double * worklin;
        if (linpred_current==1)
            worklin = linearpred1.getV();
        else
            worklin = linearpred2.getV();

        double * pmu = helpmat1.getV();

        unsigned i;
        for (i=0;i<nrobs;i++,pmu++,worklin++)
        {
            *pmu = (*worklin);
            //    double t = 0;
        }

    }

    //------------------------------------------------------------------------------
    //-------------- CLASS: DISTR_sndp_omega ---------------------------------------
    //------------------------------------------------------------------------------

    // Constructor:
    DISTR_sndp_omega::DISTR_sndp_omega(GENERAL_OPTIONS * o,
                                       const datamatrix & r,
                                       const datamatrix & w)
    : DISTR_gamlss(o,r,2,w) // <-----
    {
        family = "Skew Normal Distribution, Direct Parametrization - omega";
        outpredictor = true;
        outexpectation = false; // only true if main parameter
        predictor_name = "omega";

        linpredminlimit=-10; // necessary for log-link
        linpredmaxlimit=15; // necessary for log-link

        // updateIWLS = false; // only false and necessary if gibbs sampling?
    }


    // Copy Constructor:
    DISTR_sndp_omega::DISTR_sndp_omega(const DISTR_sndp_omega & nd)
    : DISTR_gamlss(DISTR_gamlss(nd))
    {

    }

    // Operator:
    const DISTR_sndp_omega & DISTR_sndp_omega::operator=(
                                                         const DISTR_sndp_omega & nd)
    {
        if (this==&nd)
            return *this;
        DISTR_gamlss::operator=(DISTR_gamlss(nd));
        return *this;
    }


    double DISTR_sndp_omega::get_intercept_start(void)
    {
        return 0;
    }

    void DISTR_sndp_omega::compute_param_mult(vector<double *>  linpred,double * param)
    {
        *param = exp(*linpred[1]); // <--- c.f. linpred info
    }


    double DISTR_sndp_omega::loglikelihood_weightsone(double * response,
                                                      double * linpred)
    {
        if (counter==0)
        {
            set_worklin();
        }

        double x = (*response);
        double xi = (*worklin[1]);
        double om = exp(*linpred);
        double al = (*worklin[0]);

        double z = (x - xi) / om;

        double l;

        l = - log(om) - 0.5*pow(z, 2) + log(randnumbers::Phi2(al*z)); // calculate log-likelihood (possibly simplified)

        modify_worklin();

        return l;

    }


    void DISTR_sndp_omega::compute_iwls_wweightschange_weightsone(
                                                                  double * response,
                                                                  double * linpred,
                                                                  double * workingweight,
                                                                  double * workingresponse,
                                                                  double & like,
                                                                  const bool & compute_like)
    {
        if (counter==0)
        {
            set_worklin();
        }

        double x = (*response);
        double xi = (*worklin[1]);
        double om = exp(*linpred);
        double al = (*worklin[0]);

        double z = (x - xi) / om;

        double nu = (z*z - 1 - al*randnumbers::phi(al*z) / randnumbers::Phi2(al*z)*z); // v

        *workingweight = (2 + al*al*supportfunctions::sn_a2(al)); // w

        *workingresponse = *linpred + nu/(*workingweight);

        if (compute_like)
        {

            like += - log(om) - 0.5*pow(z, 2) + log(randnumbers::Phi2(al*z)); // loglikelihood

        }


        modify_worklin();

    }



    void DISTR_sndp_omega::outoptions(void)
    {
        DISTR::outoptions();
        optionsp->out("Link function (omega): log\n");
        optionsp->out("\n");
        optionsp->out("\n");
    }


    void DISTR_sndp_omega::update_end(void)
    {


        // helpmat1 stores (eta_mu)

        double * worklin;
        if (linpred_current==1)
            worklin = linearpred1.getV();
        else
            worklin = linearpred2.getV();

        double * pmu = helpmat1.getV();

        unsigned i;
        for (i=0;i<nrobs;i++,pmu++,worklin++)
        {
            *pmu = exp(*worklin);
            //    double t = 0;
        }

    }

    //------------------------------------------------------------------------------
    //-------------- CLASS: DISTR_sndp_xi ------------------------------------------
    //------------------------------------------------------------------------------

    // ONLY FOR THE MAIN PARAMETER:
    void DISTR_sndp_xi::check_errors(void)
    {

        if (errors==false)
        {
            unsigned i=0;
            double * workresp = response.getV();
            double * workweight = weight.getV();
            while ( (i<nrobs) && (errors==false) )
            {

                if (*workweight > 0)
                {


                }
                else if (*workweight == 0)
                {
                }
                else
                {
                    errors=true;
                    errormessages.push_back("ERROR: negative weights encountered\n");
                }

                i++;
                workresp++;
                workweight++;

            }

        }

    }

    // Constructor:
    DISTR_sndp_xi::DISTR_sndp_xi(GENERAL_OPTIONS * o,
                                 const datamatrix & r,
                                 const datamatrix & w)
    : DISTR_gamlss(o,r,2,w) // <-----
    {
        family = "Skew Normal Distribution, Direct Parametrization - xi";
        outpredictor = true;
        outexpectation = true; // only true if main parameter
        predictor_name = "xi";

        // linpredminlimit=-10; // necessary for log-link
        // linpredmaxlimit=15; // necessary for log-link

        // updateIWLS = false; // only false and necessary if gibbs sampling?
    }


    // Copy Constructor:
    DISTR_sndp_xi::DISTR_sndp_xi(const DISTR_sndp_xi & nd)
    : DISTR_gamlss(DISTR_gamlss(nd))
    {

    }

    // Operator:
    const DISTR_sndp_xi & DISTR_sndp_xi::operator=(
                                                   const DISTR_sndp_xi & nd)
    {
        if (this==&nd)
            return *this;
        DISTR_gamlss::operator=(DISTR_gamlss(nd));
        return *this;
    }


    // ONLY FOR THE MAIN PARAMETER:
    void DISTR_sndp_xi::compute_deviance_mult(vector<double *> response,
                                              vector<double *> weight,
                                              vector<double *> linpred,
                                              double * deviance,
                                              vector<datamatrix*> aux)
    {
        if (*weight[2] == 0) // <---
            *deviance=0;
        else
        {

            double x = (*response[2]);
            double xi = (*linpred[2]);
            double om = exp(*linpred[1]);
            double al = (*linpred[0]);

            double z = (x - xi) / om;
            double l;

            l = -0.5*log(2*PI) - log(om) - 0.5*pow(z, 2) + log(2*randnumbers::Phi2(al*z)); // Compute log-likelihood (complete)

            *deviance = -2*l;
        }
    }


    double DISTR_sndp_xi::get_intercept_start(void)
    {
        return 0;
    }

    void DISTR_sndp_xi::compute_param_mult(vector<double *>  linpred,double * param)
    {
        *param = (*linpred[2]); // <--- c.f. linpred info
    }

    // ONLY FOR THE MAIN PARAMETER:
    double DISTR_sndp_xi::cdf_mult(vector<double *> response,
                                   vector<double *> param,
                                   vector<double *> weight,
                                   vector<datamatrix *> aux)


    {
        return 0; // calculate cdf for observation using param!
    }

    // ONLY FOR THE MAIN PARAMETER:
    double DISTR_sndp_xi::pdf_mult(vector<double *> response,
                                   vector<double *> param,
                                   vector<double *> weight,
                                   vector<datamatrix *> aux)
    {
        return 0; // usually set to be zero
    }




    double DISTR_sndp_xi::loglikelihood_weightsone(double * response,
                                                   double * linpred)
    {
        if (counter==0)
        {
            set_worklin();
        }
        double x = (*response);
        double xi = (*linpred);
        double om = exp(*worklin[1]);
        double al = (*worklin[0]);

        double z = (x - xi) / om;

        double l;

        l = - 0.5*pow(z, 2) + log(randnumbers::Phi2(al*z)); // calculate log-likelihood (possibly simplified)

        modify_worklin();

        return l;

    }


    void DISTR_sndp_xi::compute_iwls_wweightschange_weightsone(
                                                               double * response,
                                                               double * linpred,
                                                               double * workingweight,
                                                               double * workingresponse,
                                                               double & like,
                                                               const bool & compute_like)
    {
        if (counter==0)
        {
            set_worklin();
        }

        double x = (*response);
        double xi = (*linpred);
        double om = exp(*worklin[1]);
        double al = (*worklin[0]);
        double z = (x - xi)/om;

        double nu = (z - al*randnumbers::phi(al*z) / randnumbers::Phi2(al*z))/om; // v

        *workingweight = (1 + al * al * supportfunctions::sn_a0(al))/om/om; // w

        *workingresponse = *linpred + nu/(*workingweight);

        if (compute_like)
        {

            like += - 0.5*pow(z, 2) + log(randnumbers::Phi2(al*z)); // loglikelihood

        }


        modify_worklin();

    }


    // ONLY FOR THE MAIN PARAMETER:
    void DISTR_sndp_xi::compute_mu_mult(vector<double *> linpred,vector<double *> response,double * mu)
    {
        *mu = ((*linpred[predstart_mumult+2])); // <--- depends on the number of parameters
    }


    void DISTR_sndp_xi::outoptions(void)
    {
        DISTR::outoptions();
        optionsp->out("Link function (xi): identity\n");
        optionsp->out("\n");
        optionsp->out("\n");
    }


    void DISTR_sndp_xi::update_end(void)
    {


        // helpmat1 stores (eta_mu)

        double * worklin;
        if (linpred_current==1)
            worklin = linearpred1.getV();
        else
            worklin = linearpred2.getV();

        double * pmu = helpmat1.getV();

        unsigned i;
        for (i=0;i<nrobs;i++,pmu++,worklin++)
        {
            *pmu = (*worklin);
            //    double t = 0;
        }

    }

    //------------------------------------------------------------------------------
    //-------------- CLASS: DISTR_sncp_gamma ---------------------------------------
    //------------------------------------------------------------------------------

    DISTR_sncp_gamma::DISTR_sncp_gamma(GENERAL_OPTIONS * o,
                                       const datamatrix & r,
                                       const datamatrix & w)
    : DISTR_gamlss(o,r,2,w) // <-----
    {
        family = "Skew Normal Distribution, Centered Parametrization - Gamma";
        outpredictor = true;
        outexpectation = false; // only true if main parameter
        predictor_name = "gamma";

        linpredminlimit=-100; // necessary for log-link
        linpredmaxlimit=100; // necessary for log-link

        // updateIWLS = false; // only false and necessary if gibbs sampling?
    }


    // Copy Constructor:
    DISTR_sncp_gamma::DISTR_sncp_gamma(const DISTR_sncp_gamma & nd)
    : DISTR_gamlss(DISTR_gamlss(nd))
    {

    }

    // Operator:
    const DISTR_sncp_gamma & DISTR_sncp_gamma::operator=(
                                                         const DISTR_sncp_gamma & nd)
    {
        if (this==&nd)
            return *this;
        DISTR_gamlss::operator=(DISTR_gamlss(nd));
        return *this;
    }




    double DISTR_sncp_gamma::get_intercept_start(void)
    {
        return 0;
    }

    void DISTR_sncp_gamma::compute_param_mult(vector<double *>  linpred,double * param)
    {
        *param = 0.9952716*((*linpred[0])/(sqrt(1 + (*linpred[0])*(*linpred[0])))); // <--- c.f. linpred info
    }




    double DISTR_sncp_gamma::loglikelihood_weightsone(double * response,
                                                      double * linpred)
    {
        if (counter==0)
        {
            set_worklin();
        }

        double x = (*response);
        double m = *worklin[1];
        double s = exp(*worklin[0]);
        double g = 0.9952716*((*linpred)/(sqrt(1 + (*linpred)*(*linpred))));
        double R = cbrt(2*g/(4-PI));
        double b2 = 2.0/PI;
        double al = R/sqrt(b2-(1-b2)*R*R);
        double d = al/sqrt(1 + al*al);
        double om = s/sqrt(1 - b2*d*d);
        double xi = m - sqrt(b2)*om*d;

        double l = 0;

        l = - log(om) - 0.5*pow(x-xi, 2)/om/om+log(2*randnumbers::Phi2(al*(x-xi)/om)); // calculate log-likelihood (possibly simplified)

        modify_worklin();

        return l;

    }


    void DISTR_sncp_gamma::compute_iwls_wweightschange_weightsone(
                                                                  double * response,
                                                                  double * linpred,
                                                                  double * workingweight,
                                                                  double * workingresponse,
                                                                  double & like,
                                                                  const bool & compute_like)
    {
        if (counter==0)
        {
            set_worklin();
            if((*linpred) > 100.0){
              (*linpred) = 100.0;
            }
            else if((*linpred) < - 100.0){
              (*linpred) = -100.0;
            }
        }

        double x = (*response);
        double m = (*worklin[1]);
        double s = exp(*worklin[0]);
        double g = 0.9952716*((*linpred)/(sqrt(1 + (*linpred)*(*linpred))));
        double g1 = g;
        if(abs(g) < 0.001){
            g = 0.001;
        }
        double R = cbrt(2*g/(4-PI));
        double R1 = cbrt(2*g1/(4-PI));
        double b2 = 2.0/PI;
        double b = sqrt(b2);
        double al = R/sqrt(b2-(1-b2)*R*R);
        double al1 = R1/sqrt(b2-(1-b2)*R1*R1);
        double d = al/sqrt(1 + al*al);
        double d2 = d*d;
        double om = s/sqrt(1 - b2*d2);
        double xi = m - b*om*d;
        double z = (x - xi) / om;
        double zt1 = randnumbers::phi(al*z) / randnumbers::Phi2(al*z);
        double a0 = supportfunctions::sn_a0(al);
        double a1 = supportfunctions::sn_a1(al);
        double a2 = supportfunctions::sn_a2(al);

        double nu = 0;

        nu = (((al * b * zt1 - z * b + b2 * d / (1 - b2 * d2) * (al * (b * d - z) * zt1 - z * b * d - 1 + z * z))
               / pow(1 + al * al, 1.5) + z * zt1)
              * (1 / sqrt(b2 - (1 - b2) * R * R) + (1 - b2) * R * R / pow(sqrt(b2 - (1 - b2) * R * R),3)) * cbrt(2 / (4 - PI) / pow(g,2)) / 3
              * 0.9952716 / pow(sqrt(1 + (*linpred) * (*linpred)), 3));//v_g(x, m, s, g); // v
        *workingweight = ((a2 + 2 / pow(sqrt(1 + al * al), 3) * (b * al * a1 + (b * d * al * a1 - al * a2) * b2 * d / (1 - b2 * d2))
                           + (b2 * (al * al * a0 - 1) - 2 * b2 * b * d / (1 - b2 * d2) * (b * d2 * d + al * al * a1 + b * d * (1 - al * al * a0))
                              + b2 * b2 * d2 / pow(1 - b2 * d2, 2) * (2 - b2 * d2 * (1 + 2 * d2) + al * al * (a2 - 2 * b * d * a1 + b2 * d2 * a0)))
                           / pow(1 + al * al, 3))
                          * pow((1 / sqrt(b2 - (1 - b2) * R * R) + (1 - b2) * R * R / pow(sqrt(b2 - (1 - b2) * R * R),3)) * cbrt(2 / (4 - PI) / pow(g,2)) / 3
                                * 0.9952716 / pow(sqrt(1 + (*linpred) * (*linpred)), 3), 2)); //w_g(x, m, s, g); // w




        *workingresponse = *linpred + nu/(*workingweight);

        if (compute_like)
        {

            like += - log(om) - 0.5*pow(x-xi, 2)/om/om + log(2*randnumbers::Phi2(al1*(x-xi)/om)); // loglikelihood

        }


        modify_worklin();

    }



    void DISTR_sncp_gamma::outoptions(void)
    {
        DISTR::outoptions();
        optionsp->out("Link function (gamma): custom\n"); // HIER custom ersetzen
        optionsp->out("\n");
        optionsp->out("\n");
    }


    void DISTR_sncp_gamma::update_end(void)
    {


        // helpmat1 stores (eta_mu)

        double * worklin;
        if (linpred_current==1)
            worklin = linearpred1.getV();
        else
            worklin = linearpred2.getV();

        double * pmu = helpmat1.getV();

        unsigned i;
        for (i=0;i<nrobs;i++,pmu++,worklin++)
        {
            *pmu = 0.9952716*((*worklin)/(sqrt(1 + (*worklin)*(*worklin))));
            //    double t = 0;
        }

    }

    //------------------------------------------------------------------------------
    //-------------- CLASS: DISTR_sncp_sigma ---------------------------------------
    //------------------------------------------------------------------------------

    // Constructor:
    DISTR_sncp_sigma::DISTR_sncp_sigma(GENERAL_OPTIONS * o,
                                       const datamatrix & r,
                                       const datamatrix & w)
    : DISTR_gamlss(o,r,2,w) // <-----
    {
        family = "Skew Normal Distribution, Centered Parametrization - sigma";
        outpredictor = true;
        outexpectation = false; // only true if main parameter
        predictor_name = "sigma";

        linpredminlimit=-10; // necessary for log-link
        linpredmaxlimit=15; // necessary for log-link

        // updateIWLS = false; // only false and necessary if gibbs sampling?
    }


    // Copy Constructor:
    DISTR_sncp_sigma::DISTR_sncp_sigma(const DISTR_sncp_sigma & nd)
    : DISTR_gamlss(DISTR_gamlss(nd))
    {

    }

    // Operator:
    const DISTR_sncp_sigma & DISTR_sncp_sigma::operator=(
                                                         const DISTR_sncp_sigma & nd)
    {
        if (this==&nd)
            return *this;
        DISTR_gamlss::operator=(DISTR_gamlss(nd));
        return *this;
    }



    double DISTR_sncp_sigma::get_intercept_start(void)
    {
        return 0;
    }

    void DISTR_sncp_sigma::compute_param_mult(vector<double *>  linpred,double * param)
    {
        *param = exp(*linpred[1]); // <--- c.f. linpred info
    }



    double DISTR_sncp_sigma::loglikelihood_weightsone(double * response,
                                                      double * linpred)
    {
        if (counter==0)
        {
            set_worklin();
        }

        double x = (*response);
        double m = (*worklin[1]);
        double s = exp(*linpred);
        double g = 0.9952716*((*worklin[0])/(sqrt(1 + (*worklin[0])*(*worklin[0]))));
        double R = cbrt(2*g/(4-PI));
        double b2 = 2.0/PI;
        double al = R/sqrt(b2-(1-b2)*R*R);
        double d = al/sqrt(1 + al*al);
        double om = s/sqrt(1 - b2*d*d);
        double xi = m - sqrt(b2)*om*d;

        double l = 0;

        l = - log(om) - 0.5*pow(x-xi, 2)/om/om+log(2*randnumbers::Phi2(al*(x-xi)/om)); // calculate log-likelihood (possibly simplified)

        modify_worklin();

        return l;

    }


    void DISTR_sncp_sigma::compute_iwls_wweightschange_weightsone(
                                                                  double * response,
                                                                  double * linpred,
                                                                  double * workingweight,
                                                                  double * workingresponse,
                                                                  double & like,
                                                                  const bool & compute_like)
    {
        if (counter==0)
        {
            set_worklin();
        }

        double x = (*response);
        double m = (*worklin[1]);
        double s = exp(*linpred);
        double g = 0.9952716*((*worklin[0])/(sqrt(1 + (*worklin[0])*(*worklin[0]))));
        double R = cbrt(2*g/(4-PI));
        double b2 = 2.0/PI;
        double b = sqrt(b2);
        double al = R/sqrt(b2-(1-b2)*R*R);
        double d = al/sqrt(1 + al*al);
        double d2 = d*d;
        double om = s/sqrt(1 - b2*d2);
        double xi = m - b*om*d;
        double z = (x - xi) / om;

        double nu = (z * z - b * d * z - 1 - al * randnumbers::phi(al*z) / randnumbers::Phi2(al*z) * (z - b * d));//v_s(x, m, s, g);

        *workingweight = b2 * d2 * (1 + al * al * supportfunctions::sn_a0(al) - 2 * (1 + 2 * al * al) / (1 + al * al)) - 2 * b * d * al * al * supportfunctions::sn_a1(al) + 2 + al * al * supportfunctions::sn_a2(al);
        //(b2 * d * (d * (1 + al * al * sn_a0(al) - 2 * (1 - al * al) / (1 + al * al)) - 2 * al * al * sn_a1(al)) + 2 + al * al * sn_a2(al));//w_s(x, m, s, g);

        *workingresponse = *linpred + nu/(*workingweight);

        if (compute_like)
        {

            like += - log(om) - 0.5*pow(x-xi, 2)/om/om+log(2*randnumbers::Phi2(al*(x-xi)/om)); // loglikelihood

        }


        modify_worklin();

    }




    void DISTR_sncp_sigma::outoptions(void)
    {
        DISTR::outoptions();
        optionsp->out("Link function (sigma): log\n");
        optionsp->out("\n");
        optionsp->out("\n");
    }


    void DISTR_sncp_sigma::update_end(void)
    {


        // helpmat1 stores (eta_mu)

        double * worklin;
        if (linpred_current==1)
            worklin = linearpred1.getV();
        else
            worklin = linearpred2.getV();

        double * pmu = helpmat1.getV();

        unsigned i;
        for (i=0;i<nrobs;i++,pmu++,worklin++)
        {
            *pmu = exp(*worklin);
            //    double t = 0;
        }

    }

    //------------------------------------------------------------------------------
    //-------------- CLASS: DISTR_sncp_mu ------------------------------------------
    //------------------------------------------------------------------------------

    // ONLY FOR THE MAIN PARAMETER:
    void DISTR_sncp_mu::check_errors(void)
    {

        if (errors==false)
        {
            unsigned i=0;
            double * workresp = response.getV();
            double * workweight = weight.getV();
            while ( (i<nrobs) && (errors==false) )
            {

                if (*workweight > 0)
                {


                }
                else if (*workweight == 0)
                {
                }
                else
                {
                    errors=true;
                    errormessages.push_back("ERROR: negative weights encountered\n");
                }

                i++;
                workresp++;
                workweight++;

            }

        }

    }

    // Constructor:
    DISTR_sncp_mu::DISTR_sncp_mu(GENERAL_OPTIONS * o,
                                 const datamatrix & r,
                                 const datamatrix & w)
    : DISTR_gamlss(o,r,2,w) // <-----
    {
        family = "Skew Normal Distribution, Centered Parametrization - mu";
        outpredictor = true;
        outexpectation = true; // only true if main parameter
        predictor_name = "mu";

        // linpredminlimit=-10; // necessary for log-link
        // linpredmaxlimit=15; // necessary for log-link

        // updateIWLS = false; // only false and necessary if gibbs sampling?
    }


    // Copy Constructor:
    DISTR_sncp_mu::DISTR_sncp_mu(const DISTR_sncp_mu & nd)
    : DISTR_gamlss(DISTR_gamlss(nd))
    {

    }

    // Operator:
    const DISTR_sncp_mu & DISTR_sncp_mu::operator=(
                                                   const DISTR_sncp_mu & nd)
    {
        if (this==&nd)
            return *this;
        DISTR_gamlss::operator=(DISTR_gamlss(nd));
        return *this;
    }


    // ONLY FOR THE MAIN PARAMETER:
    void DISTR_sncp_mu::compute_deviance_mult(vector<double *> response,
                                              vector<double *> weight,
                                              vector<double *> linpred,
                                              double * deviance,
                                              vector<datamatrix*> aux)
    {
        if (*weight[2] == 0) // <---
            *deviance=0;
        else
        {
            double x = (*response[2]);
            double m = (*linpred[2]);
            double s = exp(*linpred[1]);
            double g = 0.9952716*((*linpred[0])/(sqrt(1 + (*linpred[0])*(*linpred[0]))));
            double R = cbrt(2*g/(4-PI));
            double b2 = 2.0/PI;
            double al = R/sqrt(b2-(1-b2)*R*R);
            double d = al/sqrt(1 + al*al);
            double om = s/sqrt(1 - b2*d*d);
            double xi = m - sqrt(b2)*om*d;

            double l;

            l = -0.5*log(2*PI) - log(om) - 0.5*pow(x-xi, 2)/om/om+log(2*randnumbers::Phi2(al*(x-xi)/om)); // Compute log-likelihood (complete)

            *deviance = -2*l;
        }
    }


    double DISTR_sncp_mu::get_intercept_start(void)
    {
        return 0;
    }

    void DISTR_sncp_mu::compute_param_mult(vector<double *>  linpred,double * param)
    {
        *param = (*linpred[2]); // <--- c.f. linpred info
    }

    // ONLY FOR THE MAIN PARAMETER:
    double DISTR_sncp_mu::cdf_mult(vector<double *> response,
                                   vector<double *> param,
                                   vector<double *> weight,
                                   vector<datamatrix *> aux)


    {
        return 0; // calculate cdf for observation using param!
    }

    // ONLY FOR THE MAIN PARAMETER:
    double DISTR_sncp_mu::pdf_mult(vector<double *> response,
                                   vector<double *> param,
                                   vector<double *> weight,
                                   vector<datamatrix *> aux)
    {
        return 0; // usually set to be zero
    }




    double DISTR_sncp_mu::loglikelihood_weightsone(double * response,
                                                   double * linpred)
    {
        if (counter==0)
        {
            set_worklin();
        }
        double x = (*response);
        double m = (*linpred);
        double s = exp(*worklin[1]);
        double g = 0.9952716*((*worklin[0])/(sqrt(1 + (*worklin[0])*(*worklin[0]))));
        double R = cbrt(2*g/(4-PI));
        double b2 = 2.0/PI;
        double al = R/sqrt(b2-(1-b2)*R*R);
        double d = al/sqrt(1 + al*al);
        double om = s/sqrt(1 - b2*d*d);
        double xi = m - sqrt(b2)*om*d;

        double l;

        l = - log(om) - 0.5*pow(x-xi, 2)/om/om+log(2*randnumbers::Phi2(al*(x-xi)/om)); // calculate log-likelihood (possibly simplified)

        modify_worklin();

        return l;

    }


    void DISTR_sncp_mu::compute_iwls_wweightschange_weightsone(
                                                               double * response,
                                                               double * linpred,
                                                               double * workingweight,
                                                               double * workingresponse,
                                                               double & like,
                                                               const bool & compute_like)
    {
        if (counter==0)
        {
            set_worklin();
        }
        double x = (*response);
        double m = (*linpred);
        double s = exp(*worklin[1]);
        double g = 0.9952716*((*worklin[0])/(sqrt(1 + (*worklin[0])*(*worklin[0]))));
        double R = cbrt(2*g/(4-PI));
        double b2 = 2.0/PI;
        double al = R/sqrt(b2-(1-b2)*R*R);
        double d = al/sqrt(1 + al*al);
        double om = s/sqrt(1 - b2*d*d);
        double xi = m - sqrt(b2)*om*d;
        double z = (x - xi) / om;

        double nu = z / om - al / om * randnumbers::phi(al*z) / randnumbers::Phi2(al*z);//v_m(x, m, s, g);

        *workingweight = (1 + al * al * supportfunctions::sn_a0(al)) / om / om;//w_m(x, m, s, g);

        *workingresponse = *linpred + nu/(*workingweight);

        if (compute_like)
        {

            like += - log(om) - 0.5*pow(x-xi, 2)/om/om+log(2*randnumbers::Phi2(al*(x-xi)/om)); // loglikelihood

        }


        modify_worklin();

    }


    // ONLY FOR THE MAIN PARAMETER:
    void DISTR_sncp_mu::compute_mu_mult(vector<double *> linpred,vector<double *> response,double * mu)
    {
        *mu = ((*linpred[predstart_mumult+2])); // <--- depends on the number of parameters
    }


    void DISTR_sncp_mu::outoptions(void)
    {
        DISTR::outoptions();
        optionsp->out("Link function (mu): identitiy\n");
        optionsp->out("\n");
        optionsp->out("\n");
    }


    void DISTR_sncp_mu::update_end(void)
    {


        // helpmat1 stores (eta_mu)

        double * worklin;
        if (linpred_current==1)
            worklin = linearpred1.getV();
        else
            worklin = linearpred2.getV();

        double * pmu = helpmat1.getV();

        unsigned i;
        for (i=0;i<nrobs;i++,pmu++,worklin++)
        {
            *pmu = (*worklin);
            //    double t = 0;
        }

    }


    //------------------------------------------------------------------------------
    //-------------- CLASS: DISTR_bivsn_alpha --------------------------------------
    //------------------------------------------------------------------------------

    // Constructor:
    DISTR_bivsn_alpha::DISTR_bivsn_alpha(GENERAL_OPTIONS * o,
                                         const datamatrix & r,
                                         unsigned & p,
                                         const datamatrix & w)
    : DISTR_gamlss(o,r,6,w) // <-----
    {
        pos = p;
        family = "Bivariate Skew Normal Distribution - Alpha";
        outpredictor = true;
        outexpectation = false; // only true if main parameter
        predictor_name = "alpha";

         linpredminlimit=-25; // necessary for log-link
         linpredmaxlimit=25; // necessary for log-link

        // updateIWLS = false; // only false and necessary if gibbs sampling?
    }


    // Copy Constructor:
    DISTR_bivsn_alpha::DISTR_bivsn_alpha(const DISTR_bivsn_alpha & nd)
    : DISTR_gamlss(DISTR_gamlss(nd))
    {
        pos = nd.pos;
        response2 = nd.response2;
        response2p = nd.response2p;
    }

    // Operator:
    const DISTR_bivsn_alpha & DISTR_bivsn_alpha::operator=(
                                                           const DISTR_bivsn_alpha & nd)
    {
        if (this==&nd)
            return *this;
        DISTR_gamlss::operator=(DISTR_gamlss(nd));
        pos = nd.pos;
        response2 = nd.response2;
        response2p = nd.response2p;
        return *this;
    }


    double DISTR_bivsn_alpha::get_intercept_start(void)
    {
        return 0;
    }

    void DISTR_bivsn_alpha::compute_param_mult(vector<double *>  linpred,double * param)
    {
        *param = (*linpred[0 + pos]); // <--- c.f. linpred info
    }


    void DISTR_bivsn_alpha::set_worklin(void)
    {

        DISTR_gamlss::set_worklin();

        response2p = response2.getV();

    }



    void DISTR_bivsn_alpha::modify_worklin(void)
    {

        DISTR_gamlss::modify_worklin();

        if (counter<nrobs-1)
        {
            response2p++;
        }

    }


    double DISTR_bivsn_alpha::loglikelihood_weightsone(double * response,
                                                       double * linpred)
    {
        if (counter==0)
        {
            set_worklin();
        }

        double al2 = (*worklin[0]);
        double al1 = (*linpred);
        double r = (*worklin[1]) / sqrt(1 + pow((*worklin[1]), 2));
        double om2 = exp((*worklin[2]));
        double om1 = exp((*worklin[3]));
        double xi2 = (*worklin[4]);
        double xi1 = (*worklin[5]);

//        double r2 = r * r;

        double l = (log(randnumbers::Phi2(al1 / om1 * ((*response) - xi1) + al2 / om2 * ((*response2p) - xi2)))); // ll, simplified

        modify_worklin();

        return l;

    }


    void DISTR_bivsn_alpha::compute_iwls_wweightschange_weightsone(
                                                                   double * response,
                                                                   double * linpred,
                                                                   double * workingweight,
                                                                   double * workingresponse,
                                                                   double & like,
                                                                   const bool & compute_like)
    {
        if (counter==0)
        {
            set_worklin();
        }

        double al2 = (*worklin[0]);
        double al1 = (*linpred);
        double r = (*worklin[1]) / sqrt(1 + pow((*worklin[1]), 2));
        double om2 = exp((*worklin[2]));
        double om1 = exp((*worklin[3]));
        double xi2 = (*worklin[4]);
        double xi1 = (*worklin[5]);

        double r2 = r * r;

        double fval = randnumbers::phi(al1 / om1 * ((*response) - xi1) + al2 / om2 * ((*response2p) - xi2));
        double Fval = randnumbers::Phi2(al1 / om1 * ((*response) - xi1) + al2 / om2 * ((*response2p) - xi2));

        double nu = ((*response) - xi1) / om1 * fval / Fval; // v

        if(nu != nu){
//            cout << "nan in alpha prevented\n";
            nu = 0.0;
        } // prevent nans (might only happen if *workingweight is fixed to a small value

        double et2 = al1 * al1 + 2.0 * r * al1 * al2 + al2 * al2;
        double apl = min(sqrt(et2 / (1 + 2 * et2)), 100.0);
        double c2 = 1 / PI / sqrt(1 + 2 * et2);
        double ma = (al1 * om1 + al2 * om1 * r) / et2;
        double sia = pow(om1, 2) - pow(om1 * (al1 + al2 * r), 2) / et2;

        if(et2 == 0){
            ma = 0.0;
            sia = pow(om1, 2);
        }


        double apl2 = apl * apl;
        double apl3 = apl * apl2;
        double apl4 = apl2 * apl2;
        double eapl = exp(apl);

        double a2 = max(-10215.5788 -10214.3724 * apl -5120.2233 * apl2 -1635.8037 * apl3 -551.0844 * apl4 + 10215.5562 * eapl, 0.00001);
        double a0 = max(-1085.73618 -1087.62413 * apl -543.95530 * apl2 -174.86737 * apl3
                        -56.98689 * apl4 + 1087.73411 * eapl, 2.0);

        *workingweight = c2 / pow(om1, 2) * (ma * ma * a2 + sia * a0); // w

        *workingresponse = *linpred + nu/(*workingweight);

        if (compute_like)
        {

            like += (log(Fval)); // loglikelihood

        }

        modify_worklin();

    }

    void DISTR_bivsn_alpha::outoptions(void)
    {
        DISTR::outoptions();
        optionsp->out("  Link function (alpha): identity\n");
        optionsp->out("\n");
        optionsp->out("\n");
    }



    void DISTR_bivsn_alpha::update_end(void)
    {


        // helpmat1 stores (eta_mu)

        double * worklin;
        if (linpred_current==1)
            worklin = linearpred1.getV();
        else
            worklin = linearpred2.getV();

        double * pmu = helpmat1.getV();

        unsigned i;
        for (i=0;i<nrobs;i++,pmu++,worklin++)
        {
            *pmu = (*worklin);
            //    double t = 0;
        }

    }




    //------------------------------------------------------------------------------
    //-------------- CLASS: DISTR_bivsn_rho ----------------------------------------
    //------------------------------------------------------------------------------

    // Constructor:
    DISTR_bivsn_rho::DISTR_bivsn_rho(GENERAL_OPTIONS * o,
                                     const datamatrix & r,
                                     const datamatrix & w)
    : DISTR_gamlss(o,r,6,w) // <-----
    {
        family = "Bivariate Skew Normal Distribution - Rho";
        outpredictor = true;
        outexpectation = false; // only true if main parameter
        predictor_name = "rho";

        linpredminlimit=-100; // necessary for log-link
        linpredmaxlimit=100; // necessary for log-link

        // updateIWLS = false; // only false and necessary if gibbs sampling?
    }


    // Copy Constructor:
    DISTR_bivsn_rho::DISTR_bivsn_rho(const DISTR_bivsn_rho & nd)
    : DISTR_gamlss(DISTR_gamlss(nd))
    {
        response2 = nd.response2;
        response2p = nd.response2p;
    }

    // Operator:
    const DISTR_bivsn_rho & DISTR_bivsn_rho::operator=(
                                                       const DISTR_bivsn_rho & nd)
    {
        if (this==&nd)
            return *this;
        DISTR_gamlss::operator=(DISTR_gamlss(nd));
        response2 = nd.response2;
        response2p = nd.response2p;
        return *this;
    }

    double DISTR_bivsn_rho::get_intercept_start(void)
    {
        return 0;
    }

    void DISTR_bivsn_rho::compute_param_mult(vector<double *>  linpred,double * param)
    {
        double arg = (*linpred[2]);
        *param = arg/pow(1+pow(arg,2),0.5); // <--- c.f. linpred info
    }


    void DISTR_bivsn_rho::set_worklin(void)
    {

        DISTR_gamlss::set_worklin();

        response2p = response2.getV();

    }



    void DISTR_bivsn_rho::modify_worklin(void)
    {

        DISTR_gamlss::modify_worklin();

        if (counter<nrobs-1)
        {
            response2p++;
        }

    }

    double DISTR_bivsn_rho::loglikelihood_weightsone(double * response,
                                                     double * linpred)
    {
        if (counter==0)
        {
            set_worklin();
        }

        double r = 0;

        if (*linpred <= -100)
            r  = -0.99995;
        else if (*linpred >= 100)
            r  = 0.99995;
        else
            r = (*linpred)/pow((1+pow((*linpred),2)),0.5);

        double r2 = r * r;
        double al2 = (*worklin[0]);
        double al1 = (*worklin[1]);
        double om2 = exp((*worklin[2]));
        double om1 = exp((*worklin[3]));
        double xi2 = (*worklin[4]);
        double xi1 = (*worklin[5]);

        double l = (log(2) - log(2 * PI) - 0.5 * log(pow(om1 * om2, 2) * (1 - r2))
                    - 0.5 * (pow(om2 * ((*response) - xi1), 2) - 2 * om1 * om2 * r * ((*response) - xi1) * ((*response2p) - xi2) + pow(om1 * ((*response2p) - xi2) , 2))
                    / (pow(om1 * om2, 2) * (1 - r2))
                    + log(randnumbers::Phi2(al1 / om1 * ((*response) - xi1) + al2 / om2 * ((*response2p) - xi2)))); // calculate log-likelihood (possibly simplified)

        modify_worklin();

        return l;

    }


    void DISTR_bivsn_rho::compute_iwls_wweightschange_weightsone(
                                                                 double * response,
                                                                 double * linpred,
                                                                 double * workingweight,
                                                                 double * workingresponse,
                                                                 double & like,
                                                                 const bool & compute_like)
    {

        if (counter==0)
        {
            set_worklin();
        }

        double r = 0;

        if (*linpred <= -100)
            r  = -0.99995;
        else if (*linpred >= 100)
            r  = 0.99995;
        else
            r = (*linpred)/pow((1+pow((*linpred),2)),0.5);

        double r2 = r * r;
        double al2 = (*worklin[0]);
        double al1 = (*worklin[1]);
        double om2 = exp((*worklin[2]));
        double om1 = exp((*worklin[3]));
        double xi2 = (*worklin[4]);
        double xi1 = (*worklin[5]);

        double dlinkf = 1.0 / sqrt(1.0 + pow(max(-100.0, min(100.0, (*linpred))), 2)) * (1 - r2);

        double nu = (r / (1 - r2) * (1 - (pow(om2 * ((*response) - xi1), 2) - 2 * om1 * om2 * r * ((*response) - xi1) * ((*response2p) - xi2) + pow(om1 * ((*response2p) - xi2) , 2))
                                     / (pow(om1 * om2, 2) * (1 - r2))) + ((*response) - xi1) * ((*response2p) - xi2) / (om1 * om2 * (1 - r2))) * dlinkf; // v

        if(nu != nu){
            nu = 0;
        } // prevent nans (might only happen if *workingweight is fixed to a small value

        *workingweight = (1 / (1 - r2) + 2 * r2 / pow(1 - r2, 2)) * pow(dlinkf, 2); // w

        *workingresponse = *linpred + nu/(*workingweight);

        if (compute_like)
        {

            like += (log(2) - log(2 * PI) - 0.5 * log(pow(om1 * om2, 2) * (1 - r2))
                     - 0.5 * (pow(om2 * ((*response) - xi1), 2) - 2 * om1 * om2 * r * ((*response) - xi1) * ((*response2p) - xi2) + pow(om1 * ((*response2p) - xi2) , 2))
                     / (pow(om1 * om2, 2) * (1 - r2))
                     + log(randnumbers::Phi2(al1 / om1 * ((*response) - xi1) + al2 / om2 * ((*response2p) - xi2))));

        }


        modify_worklin();

    }

    void DISTR_bivsn_rho::outoptions(void)
    {
        DISTR::outoptions();
        optionsp->out("  Response function (rho): rhogit\n");
        optionsp->out("\n");
        optionsp->out("\n");
    }


    void DISTR_bivsn_rho::update_end(void)
    {


        // helpmat1 stores (eta_mu)

        double * worklin;
        if (linpred_current==1)
            worklin = linearpred1.getV();
        else
            worklin = linearpred2.getV();

        double * pmu = helpmat1.getV();

        unsigned i;
        for (i=0;i<nrobs;i++,pmu++,worklin++)
        {
            *pmu = (*worklin)/pow((1+pow((*worklin),2)),0.5);
            //    double t = 0;
        }

    }


    //------------------------------------------------------------------------------
    //-------------- CLASS: DISTR_bivsn_omega --------------------------------------
    //------------------------------------------------------------------------------

    // Constructor:
    DISTR_bivsn_omega::DISTR_bivsn_omega(GENERAL_OPTIONS * o,
                                         const datamatrix & r,
                                         unsigned &p,
                                         const datamatrix & w)
    : DISTR_gamlss(o,r,6,w) // <-----
    {
        pos = p;
        family = "Bivariate Skew Normal Distribution - Omega";
        outpredictor = true;
        outexpectation = false; // only true if main parameter
        predictor_name = "omega";

        linpredminlimit=-10; // necessary for log-link
        linpredmaxlimit=15; // necessary for log-link

        // updateIWLS = false; // only false and necessary if gibbs sampling?
    }


    // Copy Constructor:
    DISTR_bivsn_omega::DISTR_bivsn_omega(const DISTR_bivsn_omega & nd)
    : DISTR_gamlss(DISTR_gamlss(nd))
    {
        pos = nd.pos;
        response2 = nd.response2;
        response2p = nd.response2p;
    }

    // Operator:
    const DISTR_bivsn_omega & DISTR_bivsn_omega::operator=(
                                                           const DISTR_bivsn_omega & nd)
    {
        if (this==&nd)
            return *this;
        DISTR_gamlss::operator=(DISTR_gamlss(nd));
        pos = nd.pos;
        response2 = nd.response2;
        response2p = nd.response2p;
        return *this;
    }

    double DISTR_bivsn_omega::get_intercept_start(void)
    {
        return 0;
    }

    void DISTR_bivsn_omega::compute_param_mult(vector<double *>  linpred,double * param)
    {
        *param = exp(*linpred[3 + pos]); // <--- c.f. linpred info
    }

    void DISTR_bivsn_omega::set_worklin(void)
    {

        DISTR_gamlss::set_worklin();

        response2p = response2.getV();

    }



    void DISTR_bivsn_omega::modify_worklin(void)
    {

        DISTR_gamlss::modify_worklin();

        if (counter<nrobs-1)
        {
            response2p++;
        }

    }

    double DISTR_bivsn_omega::loglikelihood_weightsone(double * response,
                                                       double * linpred)
    {
        if (counter==0)
        {
            set_worklin();
        }
        double al2 = (*worklin[0]);
        double al1 = (*worklin[1]);
        double r = (*worklin[2]) / sqrt(1 + pow((*worklin[2]), 2));
        double r2 = r*r;
        double om2 = exp((*worklin[3]));
        double om1 = exp((*linpred));
        double xi2 = (*worklin[4]);
        double xi1 = (*worklin[5]);


        double l;

        l = (log(2) - log(2 * PI) - 0.5 * log(pow(om1 * om2, 2) * (1 - r2))
             - 0.5 * (pow(om2 * ((*response) - xi1), 2) - 2 * om1 * om2 * r * ((*response) - xi1) * ((*response2p) - xi2) + pow(om1 * ((*response2p) - xi2) , 2))
             / (pow(om1 * om2, 2) * (1 - r2))
             + log(randnumbers::Phi2(al1 / om1 * ((*response) - xi1) + al2 / om2 * ((*response2p) - xi2)))); // calculate log-likelihood (possibly simplified)

        modify_worklin();

        return l;

    }


    void DISTR_bivsn_omega::compute_iwls_wweightschange_weightsone(
                                                                   double * response,
                                                                   double * linpred,
                                                                   double * workingweight,
                                                                   double * workingresponse,
                                                                   double & like,
                                                                   const bool & compute_like)
    {
        if (counter==0)
        {
            set_worklin();
        }

        double al2 = (*worklin[0]);
        double al1 = (*worklin[1]);
        double r = (*worklin[2]) / sqrt(1 + pow((*worklin[2]), 2));
        double r2 = r * r;
        double om2 = exp((*worklin[3]));
        double om1 = exp((*linpred));
        double xi2 = (*worklin[4]);
        double xi1 = (*worklin[5]);

        double nu = ( - 1 + pow(((*response) - xi1), 2) / (pow(om1,2) * (1 - r2)) - r * ((*response) - xi1) * ((*response2p) - xi2) / (om1 * om2 * (1 - r2))
                     - al1 *  ((*response) - xi1) / om1 * randnumbers::phi(al1 / om1 * ((*response) - xi1) + al2 / om2 * ((*response2p) - xi2))
                     / randnumbers::Phi2(al1 / om1 * ((*response) - xi1) + al2 / om2 * ((*response2p) - xi2))); // v

        if(nu != nu){
            //cout << "nu is nan in omega\n";
            nu = 0;} // prevent nans (might only happen if *workingweight is fixed to a small value

        double et2 = al1 * al1 + 2.0 * r * al1 * al2 + al2 * al2;
        double apl = sqrt(et2 / (1 + 2 * et2));
        double c2 = 1 / PI / sqrt(1 + 2 * et2);
        double ma = (al1 * om1 + al2 * om1 * r) / et2;
        double sia = pow(om1, 2) - pow(om1 * (al1 + al2 * r), 2) / et2;

        if(et2 == 0){
            ma = 0.0;
            sia = pow(om1, 2);
        }

        double apl2 = apl * apl;
        double apl3 = apl * apl2;
        double apl4 = apl2 * apl2;
        double eapl = exp(apl);

        double a2 = max(-10215.5788 -10214.3724 * apl -5120.2233 * apl2 -1635.8037 * apl3 -551.0844 * apl4 + 10215.5562 * eapl, 0.00001);
        double a0 = max(-1085.73618 -1087.62413 * apl -543.95530 * apl2 -174.86737 * apl3
                        -56.98689 * apl4 + 1087.73411 * eapl, 2.0);

        *workingweight = (3 - 2 * r2) / (1 - r2) - 1 + pow(al1 / om1, 2) * c2 * (ma * ma * a2 + sia * a0); // w

        *workingresponse = *linpred + nu/(*workingweight);

        if (compute_like)
        {

            like += (log(2) - log(2 * PI) - 0.5 * log(pow(om1 * om2, 2) * (1 - r2))
                     - 0.5 * (pow(om2 * ((*response) - xi1), 2) - 2 * om1 * om2 * r * ((*response) - xi1) * ((*response2p) - xi2) + pow(om1 * ((*response2p) - xi2) , 2))
                     / (pow(om1 * om2, 2) * (1 - r2))
                     + log(randnumbers::Phi2(al1 / om1 * ((*response) - xi1) + al2 / om2 * ((*response2p) - xi2)))); // loglikelihood

        }


        modify_worklin();

    }

    void DISTR_bivsn_omega::outoptions(void)
    {
        DISTR::outoptions();
        optionsp->out("  Link function (omega): log\n");
        optionsp->out("\n");
        optionsp->out("\n");
    }


    void DISTR_bivsn_omega::update_end(void)
    {


        // helpmat1 stores (eta_mu)

        double * worklin;
        if (linpred_current==1)
            worklin = linearpred1.getV();
        else
            worklin = linearpred2.getV();

        double * pmu = helpmat1.getV();

        unsigned i;
        for (i=0;i<nrobs;i++,pmu++,worklin++)
        {
            *pmu = exp(*worklin);
            //    double t = 0;
        }

    }


    //------------------------------------------------------------------------------
    //-------------- CLASS: DISTR_bivsn_xi -----------------------------------------
    //------------------------------------------------------------------------------

    // ONLY FOR THE MAIN PARAMETER:
    void DISTR_bivsn_xi::check_errors(void)
    {

        if (errors==false)
        {
            unsigned i=0;
            double * workresp = response.getV();
            double * workweight = weight.getV();
            while ( (i<nrobs) && (errors==false) )
            {

                if (*workweight > 0)
                {


                }
                else if (*workweight == 0)
                {
                }
                else
                {
                    errors=true;
                    errormessages.push_back("ERROR: negative weights encountered\n");
                }

                i++;
                workresp++;
                workweight++;

            }

        }

    }

    // Constructor:
    DISTR_bivsn_xi::DISTR_bivsn_xi(GENERAL_OPTIONS * o,
                                   const datamatrix & r,
                                   unsigned &p,
                                   const datamatrix & w)
    : DISTR_gamlss(o,r,6,w) // <-----
    {
        pos = p;
        family = "Bivariate Skew Normal Distribution - Xi";
        outpredictor = true;
        outexpectation = true; // only true if main parameter
        predictor_name = "xi";

        // linpredminlimit=-10; // necessary for log-link
        // linpredmaxlimit=15; // necessary for log-link

        // updateIWLS = false; // only false and necessary if gibbs sampling?
    }


    // Copy Constructor:
    DISTR_bivsn_xi::DISTR_bivsn_xi(const DISTR_bivsn_xi & nd)
    : DISTR_gamlss(DISTR_gamlss(nd))
    {
        pos = nd.pos;
        response2 = nd.response2;
        response2p = nd.response2p;
    }

    // Operator:
    const DISTR_bivsn_xi & DISTR_bivsn_xi::operator=(
                                                     const DISTR_bivsn_xi & nd)
    {
        if (this==&nd)
            return *this;
        DISTR_gamlss::operator=(DISTR_gamlss(nd));
        pos = nd.pos;
        response2 = nd.response2;
        response2p = nd.response2p;
        return *this;
    }


    // ONLY FOR THE MAIN PARAMETER:
    void DISTR_bivsn_xi::compute_deviance_mult(vector<double *> response,
                                               vector<double *> weight,
                                               vector<double *> linpred,
                                               double * deviance,
                                               vector<datamatrix*> aux)
    {
        if (*weight[6] == 0) // <---
            *deviance=0;
        else
        {
            double al2 = (*linpred[0]);
            double al1 = (*linpred[1]);
            double r = (*linpred[2]) / sqrt(1 + pow((*linpred[2]), 2));
            double r2 = r * r;
            double om2 = exp((*linpred[3]));
            double om1 = exp((*linpred[4]));
            double xi2 = (*linpred[5]);
            double xi1 = (*linpred[6]);

            double l = 0;

            l = (log(2) - log(2 * PI) - 0.5 * (log(pow(om1 * om2, 2) * (1 - r2))
                                               + (pow(om2 * ((*response[6]) - xi1), 2) - 2 * om1 * om2 * r * ((*response[6]) - xi1) * ((*response[5]) - xi2) + pow(om1 * ((*response[5]) - xi2) , 2))
                                               / (pow(om1 * om2, 2) * (1 - r2)))
                 + log(randnumbers::Phi2(al1 / om1 * ((*response[6]) - xi1) + al2 / om2 * ((*response[5]) - xi2)))); // Compute log-likelihood (complete) HERE sind 0 und 1 richtig oder anders

            *deviance = -2*l;
        }
    }


    double DISTR_bivsn_xi::get_intercept_start(void)
    {
        return 0;
    }

    void DISTR_bivsn_xi::compute_param_mult(vector<double *>  linpred,double * param)
    {
        *param = (*linpred[5+pos]); // <--- c.f. linpred info
    }

    // ONLY FOR THE MAIN PARAMETER:
    double DISTR_bivsn_xi::cdf_mult(vector<double *> response,
                                    vector<double *> param,
                                    vector<double *> weight,
                                    vector<datamatrix *> aux)


    {
        return 0; // calculate cdf for observation using param!
    }

    // ONLY FOR THE MAIN PARAMETER:
    double DISTR_bivsn_xi::pdf_mult(vector<double *> response,
                                    vector<double *> param,
                                    vector<double *> weight,
                                    vector<datamatrix *> aux)
    {
        return 0; // usually set to be zero
    }

    void DISTR_bivsn_xi::set_worklin(void)
    {

        DISTR_gamlss::set_worklin();

        response2p = response2.getV();

    }



    void DISTR_bivsn_xi::modify_worklin(void)
    {

        DISTR_gamlss::modify_worklin();

        if (counter<nrobs-1)
        {
            response2p++;
        }

    }


    double DISTR_bivsn_xi::loglikelihood_weightsone(double * response,
                                                    double * linpred)
    {
        if (counter==0)
        {
            set_worklin();
        }

        double al2 = (*worklin[0]);
        double al1 = (*worklin[1]);
        double r = (*worklin[2]) / sqrt(1 + pow((*worklin[2]), 2));
        double r2 = r * r;
        double om2 = exp((*worklin[3]));
        double om1 = exp((*worklin[4]));
        double xi2 = (*worklin[5]);
        double xi1 = (*linpred);

        double l = 0;

        l = (log(2) - log(2 * PI) - 0.5 * log(pow(om1 * om2, 2) * (1 - r2))
             - 0.5 * (pow(om2 * ((*response) - xi1), 2) - 2 * om1 * om2 * r * ((*response) - xi1) * ((*response2p) - xi2) + pow(om1 * ((*response2p) - xi2) , 2))
             / (pow(om1 * om2, 2) * (1 - r2))
             + log(randnumbers::Phi2(al1 / om1 * ((*response) - xi1) + al2 / om2 * ((*response2p) - xi2)))); // calculate log-likelihood (possibly simplifie

        modify_worklin();

        return l;

    }


    void DISTR_bivsn_xi::compute_iwls_wweightschange_weightsone(
                                                                double * response,
                                                                double * linpred,
                                                                double * workingweight,
                                                                double * workingresponse,
                                                                double & like,
                                                                const bool & compute_like)
    {
        if (counter==0)
        {
            set_worklin();
        }

        double al2 = (*worklin[0]);
        double al1 = (*worklin[1]);
        double r = (*worklin[2]) / sqrt(1 + pow((*worklin[2]), 2));
        double r2 = r * r;
        double om2 = exp((*worklin[3]));
        double om1 = exp((*worklin[4]));
        double xi2 = (*worklin[5]);
        double xi1 = (*linpred);

        double fval = randnumbers::phi(al1 / om1 * ((*response) - xi1) + al2 / om2 * ((*response2p) - xi2));
        double Fval = randnumbers::Phi2(al1 / om1 * ((*response) - xi1) + al2 / om2 * ((*response2p) - xi2));

        double nu = (((*response) - xi1) / (om1 * (1 - r2)) - r * ((*response2p) - xi2) / om2 / (1 - r2) - al1 * fval / Fval) / om1;


        if(nu != nu){
            nu = 0.0;
            //cout << "nan in xi prevented\n";
        } // prevent nans (might only happen if *workingweight is fixed to a small value

        double et2 = al1 * al1 + 2.0 * r * al1 * al2 + al2 * al2;
        double apl = sqrt(et2 / (1 + 2 * et2));
        double c2 = 1 / PI / sqrt(1 + 2 * et2);

        double apl2 = apl * apl;
        double apl3 = apl * apl2;
        double apl4 = apl2 * apl2;
        double eapl = exp(apl);

        double a0 = max(-1085.73618 -1087.62413 * apl -543.95530 * apl2 -174.86737 * apl3
                        -56.98689 * apl4 + 1087.73411 * eapl, 0.0);

        *workingweight = 1.0 / (pow(om1, 2) * (1 - r2)) + pow(al1, 2) * c2 * a0 / pow(om1, 2); // w

        *workingresponse = *linpred + nu/(*workingweight);

        if (compute_like)
        {

            like += (log(2) - log(2 * PI) - 0.5 * log(pow(om1 * om2, 2) * (1 - r2))
                     - 0.5 * (pow(om2 * ((*response) - xi1), 2) - 2 * om1 * om2 * r * ((*response) - xi1) * ((*response2p) - xi2) + pow(om1 * ((*response2p) - xi2) , 2))
                     / (pow(om1 * om2, 2) * (1 - r2))
                     + log(Fval)); // loglikelihood

        }


        modify_worklin();

    }


    // ONLY FOR THE MAIN PARAMETER:
    void DISTR_bivsn_xi::compute_mu_mult(vector<double *> linpred,vector<double *> response,double * mu)
    {
        *mu = ((*linpred[predstart_mumult+5+pos])); // <--- depends on the number of parameters
    }


    void DISTR_bivsn_xi::outoptions(void)
    {
        DISTR::outoptions();
        optionsp->out("  Response function (xi): identity\n");
        optionsp->out("\n");
        optionsp->out("\n");
    }


    void DISTR_bivsn_xi::update_end(void)
    {


        // helpmat1 stores (eta_mu)

        double * worklin;
        if (linpred_current==1)
            worklin = linearpred1.getV();
        else
            worklin = linearpred2.getV();

        double * pmu = helpmat1.getV();

        unsigned i;
        for (i=0;i<nrobs;i++,pmu++,worklin++)
        {
            *pmu = (*worklin);
            //    double t = 0;:
        }
    }


    //------------------------------------------------------------------------------
    //-------------- CLASS: DISTR_st_df --------------------------------------------
    //------------------------------------------------------------------------------

    // Constructor:
    DISTR_st_df::DISTR_st_df(GENERAL_OPTIONS * o,
                             const datamatrix & r,
                             const datamatrix & w)
    : DISTR_gamlss(o,r,3,w)
    {
        family = "Skew t Distribution Direct Parameterization - df";
        outpredictor = true;
        outexpectation = false;
        predictor_name = "df";

        linpredminlimit=-3;
        linpredmaxlimit=7;

        STOREDVALUES::get_st_ef_df_values(ef);
        STOREDVALUES::get_st_iota_values(io);
        // updateIWLS = false; // only false and necessary if gibbs sampling?
    }


    // Copy Constructor:
    DISTR_st_df::DISTR_st_df(const DISTR_st_df & nd)
    : DISTR_gamlss(DISTR_gamlss(nd))
    {
      STOREDVALUES::get_st_ef_df_values(ef);
      STOREDVALUES::get_st_iota_values(io);
    }

    // Operator:
    const DISTR_st_df & DISTR_st_df::operator=(
                                               const DISTR_st_df & nd)
    {
        if (this==&nd)
            return *this;
        DISTR_gamlss::operator=(DISTR_gamlss(nd));
        STOREDVALUES::get_st_ef_df_values(ef);
        STOREDVALUES::get_st_iota_values(io);
        return *this;
    }


    double DISTR_st_df::get_intercept_start(void)
    {
        return 0;
    }

    void DISTR_st_df::compute_param_mult(vector<double *>  linpred,double * param)
    {
        *param = exp(*linpred[0]); // <--- c.f. linpred info
    }


    double DISTR_st_df::loglikelihood_weightsone(double * response,
                                                 double * linpred)
    {
        if (counter==0)
        {
            set_worklin();
        }

        double x = (*response);
        double xi = (*worklin[2]);
        double omega = exp(*worklin[1]);
        double alpha = (*worklin[0]);
        double df = exp(*linpred);
        double dfplusonehalf = 0.5*(df+1.0);
        double dfhalf = 0.5 * df;
        double z = (x - xi) / omega;
        double tau = sqrt((df+1)/(pow(z,2) + df));

        double l;

        l = - 0.5*log(df) + randnumbers::lngamma_exact(dfplusonehalf) - randnumbers::lngamma_exact(dfhalf) - dfplusonehalf*log(1 + pow(z,2)/df) + log(gsl_cdf_tdist_P(alpha*z*tau, max(30.0, df+1.0)));

        modify_worklin();

        return l;

    }


    void DISTR_st_df::compute_iwls_wweightschange_weightsone(
                                                             double * response,
                                                             double * linpred,
                                                             double * workingweight,
                                                             double * workingresponse,
                                                             double & like,
                                                             const bool & compute_like)
    {
        if (counter==0)
        {
            set_worklin();
        }

        double x = (*response);
        double xi = (*worklin[2]);
        double omega = exp(*worklin[1]);
        double alpha = (*worklin[0]);
        double df = exp(*linpred);
        double dfplusonehalf = 0.5*(df+1.0);
        double dfhalfplusone = 0.5*df + 1.0;
        double dfhalf = 0.5 * df;
        double z = (x - xi) / omega;
        double tau = sqrt((df+1.0)/(pow(z,2) + df));

        double tmp = alpha*z*tau;
        double tval = gsl_ran_tdist_pdf(tmp, max(30.0, df + 1.0));
        double Tval = gsl_cdf_tdist_P(tmp, max(30.0, df+1.0));
        double iota = io[(max(-100, min(100, (int)nearbyint(tmp*2.0))) + 100)][(max(-20, min(50, (int)nearbyint((*linpred)*10.0))) + 20)];

        double nu = 0.5 * (randnumbers::digamma_exact(dfhalfplusone) - randnumbers::digamma_exact(dfhalf) - (2 * df + 1) / (df + 1) / df - log(1.0 + pow(z,2) / df)
                          + pow(z * tau, 2) / df + alpha * z * (z * z - 1) / pow(df+z*z, 2) / tau * tval / Tval + iota / Tval) * df; // v

        *workingweight = ef[(max(-200, min(500, (int)nearbyint((*linpred)*100.0))) + 200)];// * pow(df, 2); // w

        *workingresponse = *linpred + nu/(*workingweight);

        if (compute_like)
        {

            like += - 0.5*log(df) + randnumbers::lngamma_exact(dfplusonehalf) - randnumbers::lngamma_exact(dfhalf) - 0.5*(df+1)*log(1 + pow(z,2)/df) + log(Tval); // loglikelihood

        }


        modify_worklin();

    }

    void DISTR_st_df::outoptions(void)
    {
        DISTR::outoptions();
        optionsp->out("Link function (df): log\n");
        optionsp->out("\n");
        optionsp->out("\n");
    }


    void DISTR_st_df::update_end(void)
    {


        // helpmat1 stores (eta_mu)

        double * worklin;
        if (linpred_current==1)
            worklin = linearpred1.getV();
        else
            worklin = linearpred2.getV();

        double * pmu = helpmat1.getV();

        unsigned i;
        for (i=0;i<nrobs;i++,pmu++,worklin++)
        {
            *pmu = exp(*worklin);
        }

    }

    //------------------------------------------------------------------------------
    //-------------- CLASS: DISTR_st_alpha -----------------------------------------
    //------------------------------------------------------------------------------

    // Constructor:
    DISTR_st_alpha::DISTR_st_alpha(GENERAL_OPTIONS * o,
                                   const datamatrix & r,
                                   const datamatrix & w)
    : DISTR_gamlss(o,r,3,w)
    {
        family = "Skew t Distribution - alpha";
        outpredictor = true;
        outexpectation = false; // only true if main parameter
        predictor_name = "alpha";

        linpredminlimit=-50; // necessary for log-link
        linpredmaxlimit=50; // necessary for log-link

        STOREDVALUES::get_st_ef_alpha_values(ef);
        // updateIWLS = false; // only false and necessary if gibbs sampling?
    }


    // Copy Constructor:
    DISTR_st_alpha::DISTR_st_alpha(const DISTR_st_alpha & nd)
    : DISTR_gamlss(DISTR_gamlss(nd))
    {
      STOREDVALUES::get_st_ef_alpha_values(ef);
    }

    // Operator:
    const DISTR_st_alpha & DISTR_st_alpha::operator=(
                                                     const DISTR_st_alpha & nd)
    {
        if (this==&nd)
            return *this;
        DISTR_gamlss::operator=(DISTR_gamlss(nd));
        STOREDVALUES::get_st_ef_alpha_values(ef);
        return *this;
    }

    double DISTR_st_alpha::get_intercept_start(void)
    {
        return 0;
    }

    void DISTR_st_alpha::compute_param_mult(vector<double *>  linpred,double * param)
    {
        *param = (*linpred[1]); // <--- c.f. linpred info
    }

    double DISTR_st_alpha::loglikelihood_weightsone(double * response,
                                                    double * linpred)
    {
        if (counter==0)
        {
            set_worklin();
        }

        double x = (*response);
        double xi = *worklin[2];
        double omega = exp(*worklin[1]);
        double alpha = *linpred;
        double df = exp(*worklin[0]);
        double z = (x - xi) / omega;
        double tau = sqrt((df+1)/(pow(z,2) + df));

        double l;

        l = log(gsl_cdf_tdist_P(alpha*z*tau, max(30.0, df+1))); // calculate log-likelihood (possibly simplified)

        modify_worklin();

        return l;

    }


    void DISTR_st_alpha::compute_iwls_wweightschange_weightsone(
                                                                double * response,
                                                                double * linpred,
                                                                double * workingweight,
                                                                double * workingresponse,
                                                                double & like,
                                                                const bool & compute_like)
    {
        if (counter==0)
        {
            set_worklin();
        }

        double x = (*response);
        double xi = *worklin[2];
        double omega = exp(*worklin[1]);
        double alpha = *linpred;
        double df = exp(*worklin[0]);
        double z = (x - xi) / omega;
        double tau = sqrt((df+1)/(pow(z,2) + df));

        double nu = z * tau * gsl_ran_tdist_pdf(alpha*z*tau, max(30.0,df + 1)) / gsl_cdf_tdist_P(alpha*z*tau, max(30.0, df + 1)); // v

        *workingweight = ef[(max(-100, min(100, (int)nearbyint((*linpred)*10.0))) + 100)][(max(-20, min(50, (int)nearbyint((*worklin[0])*10.0))) + 20)]; // w

        *workingresponse = *linpred + nu/(*workingweight);

        if (compute_like)
        {

            like += log(gsl_cdf_tdist_P(alpha*z*tau, max(30.0, df+1))); // loglikelihood

        }


        modify_worklin();

    }


    void DISTR_st_alpha::outoptions(void)
    {
        DISTR::outoptions();
        optionsp->out("Link function (alpha): identity\n");
        optionsp->out("\n");
        optionsp->out("\n");
    }


    void DISTR_st_alpha::update_end(void)
    {


        // helpmat1 stores (eta_mu)

        double * worklin;
        if (linpred_current==1)
            worklin = linearpred1.getV();
        else
            worklin = linearpred2.getV();

        double * pmu = helpmat1.getV();

        unsigned i;
        for (i=0;i<nrobs;i++,pmu++,worklin++)
        {
            *pmu = (*worklin);
            //    double t = 0;
        }

    }

    //------------------------------------------------------------------------------
    //-------------- CLASS: DISTR_st_omega -----------------------------------------
    //------------------------------------------------------------------------------

    // Constructor:
    DISTR_st_omega::DISTR_st_omega(GENERAL_OPTIONS * o,
                                   const datamatrix & r,
                                   const datamatrix & w)
    : DISTR_gamlss(o,r,3,w) // <-----
    {
        family = "Skew t Distribution - omega";
        outpredictor = true;
        outexpectation = false; // only true if main parameter
        predictor_name = "omega";

        linpredminlimit=-10; // necessary for log-link
        linpredmaxlimit=15; // necessary for log-link

        STOREDVALUES::get_st_ef_omega_values(ef);
        // updateIWLS = false; // only false and necessary if gibbs sampling?
    }


    // Copy Constructor:
    DISTR_st_omega::DISTR_st_omega(const DISTR_st_omega & nd)
    : DISTR_gamlss(DISTR_gamlss(nd))
    {
      STOREDVALUES::get_st_ef_omega_values(ef);
    }

    // Operator:
    const DISTR_st_omega & DISTR_st_omega::operator=(
                                                     const DISTR_st_omega & nd)
    {
        if (this==&nd)
            return *this;
        DISTR_gamlss::operator=(DISTR_gamlss(nd));
        STOREDVALUES::get_st_ef_omega_values(ef);
        return *this;
    }

    double DISTR_st_omega::get_intercept_start(void)
    {
        return 0;
    }

    void DISTR_st_omega::compute_param_mult(vector<double *>  linpred,double * param)
    {
        *param = exp(*linpred[2]); // <--- c.f. linpred info
    }


    double DISTR_st_omega::loglikelihood_weightsone(double * response,
                                                    double * linpred)
    {
        if (counter==0)
        {
            set_worklin();
        }
        double x = (*response);
        double xi = (*worklin[2]);
        double omega = exp(*linpred);
        double alpha = *worklin[1];
        double df = exp(*worklin[0]);
        double z = (x - xi) / omega;
        double tau = sqrt((df+1)/(pow(z,2) + df));
        double l;

        l = - log(omega) - 0.5*(df+1)*log(1 + pow(z,2)/df) + log(gsl_cdf_tdist_P(alpha*z*tau, max(30.0, df+1))); // calculate log-likelihood (possibly simplified)

        modify_worklin();

        return l;

    }


    void DISTR_st_omega::compute_iwls_wweightschange_weightsone(
                                                                double * response,
                                                                double * linpred,
                                                                double * workingweight,
                                                                double * workingresponse,
                                                                double & like,
                                                                const bool & compute_like)
    {
        if (counter==0)
        {
            set_worklin();
        }
        double x = (*response);
        double xi = (*worklin[2]);
        double omega = exp(*linpred);
        double alpha = *worklin[1];
        double df = exp(*worklin[0]);
        double z = (x - xi) / omega;
        double tau = sqrt((df+1)/(pow(z,2) + df));

        double nu = (- 1 + pow(z*tau, 2) - alpha*z*tau*df / (pow(z,2) + df) * gsl_ran_tdist_pdf(alpha*z*tau, max(30.0, df + 1.0)) / gsl_cdf_tdist_P(alpha*z*tau, max(30.0, df + 1.0))); // v

        *workingweight = (ef[(max(-100, min(100, (int)nearbyint((*worklin[1])*10.0))) + 100)][(max(-20, min(50, (int)nearbyint((*worklin[0])*10.0))) + 20)]);  // w

        *workingresponse = *linpred + nu/(*workingweight);

        if (compute_like)
        {

            like += - log(omega) - 0.5*(df+1)*log(1 + pow(z,2)/df) + log(gsl_cdf_tdist_P(alpha*z*tau, max(30.0, df+1))); // loglikelihood

        }


        modify_worklin();

    }



    void DISTR_st_omega::outoptions(void)
    {
        DISTR::outoptions();
        optionsp->out("Link function (omega): log\n");
        optionsp->out("\n");
        optionsp->out("\n");
    }


    void DISTR_st_omega::update_end(void)
    {


        // helpmat1 stores (eta_mu)

        double * worklin;
        if (linpred_current==1)
            worklin = linearpred1.getV();
        else
            worklin = linearpred2.getV();

        double * pmu = helpmat1.getV();

        unsigned i;
        for (i=0;i<nrobs;i++,pmu++,worklin++)
        {
            *pmu = exp(*worklin);
            //    double t = 0;
        }

    }


    //------------------------------------------------------------------------------
    //-------------- CLASS: DISTR_st_xi --------------------------------------------
    //------------------------------------------------------------------------------

    // ONLY FOR THE MAIN PARAMETER:
    void DISTR_st_xi::check_errors(void)
    {

        if (errors==false)
        {
            unsigned i=0;
            double * workresp = response.getV();
            double * workweight = weight.getV();
            while ( (i<nrobs) && (errors==false) )
            {

                if (*workweight > 0)
                {


                }
                else if (*workweight == 0)
                {
                }
                else
                {
                    errors=true;
                    errormessages.push_back("ERROR: negative weights encountered\n");
                }

                i++;
                workresp++;
                workweight++;

            }

        }

    }

    // Constructor:
    DISTR_st_xi::DISTR_st_xi(GENERAL_OPTIONS * o,
                             const datamatrix & r,
                             const datamatrix & w)
    : DISTR_gamlss(o,r,3,w)
    {
        family = "Skew t Distribution - xi";
        outpredictor = true;
        outexpectation = true;
        predictor_name = "xi";

        STOREDVALUES::get_st_ef_xi_values(ef);
        // linpredminlimit=-10; // necessary for log-link
        // linpredmaxlimit=15; // necessary for log-link

        // updateIWLS = false; // only false and necessary if gibbs sampling?
    }


    // Copy Constructor:
    DISTR_st_xi::DISTR_st_xi(const DISTR_st_xi & nd)
    : DISTR_gamlss(DISTR_gamlss(nd))
    {
      STOREDVALUES::get_st_ef_xi_values(ef);
    }

    // Operator:
    const DISTR_st_xi & DISTR_st_xi::operator=(
                                               const DISTR_st_xi & nd)
    {
        if (this==&nd)
            return *this;
        DISTR_gamlss::operator=(DISTR_gamlss(nd));
        STOREDVALUES::get_st_ef_xi_values(ef);
        return *this;
    }



    void DISTR_st_xi::compute_deviance_mult(vector<double *> response,
                                            vector<double *> weight,
                                            vector<double *> linpred,
                                            double * deviance,
                                            vector<datamatrix*> aux)
    {
        if (*weight[3] == 0)
            *deviance=0;
        else
        {
            double x = *response[0];
            double xi = *linpred[3];
            double omega = exp(*linpred[2]);
            double alpha = *linpred[1];
            double df = exp(*linpred[0]);
            double dfplusonehalf = 0.5*(df+1.0);
            double dfhalf = 0.5 * df;
            double z = (x - xi) / omega;
            double tau = sqrt((df+1)/(pow(z,2) + df));

            double l;

            l = log(2) - log(omega) - 0.5*log(df) - 0.5*log(PI) + randnumbers::lngamma_exact(dfplusonehalf) - randnumbers::lngamma_exact(dfhalf) - 0.5*(df+1)*log(1 + pow(z,2)/df) + log(gsl_cdf_tdist_P(alpha*z*tau, max(30.0, df+1)));

            *deviance = -2*l;
        }
    }


    double DISTR_st_xi::get_intercept_start(void)
    {
        return 0;
    }

    void DISTR_st_xi::compute_param_mult(vector<double *>  linpred,double * param)
    {
        *param = (*linpred[3]); // <--- c.f. linpred info
    }

    // ONLY FOR THE MAIN PARAMETER:
    double DISTR_st_xi::cdf_mult(vector<double *> response,
                                 vector<double *> param,
                                 vector<double *> weight,
                                 vector<datamatrix *> aux)


    {
        return 0; // calculate cdf for observation using param!
    }

    // ONLY FOR THE MAIN PARAMETER:
    double DISTR_st_xi::pdf_mult(vector<double *> response,
                                 vector<double *> param,
                                 vector<double *> weight,
                                 vector<datamatrix *> aux)
    {
        return 0; // usually set to be zero
    }




    double DISTR_st_xi::loglikelihood_weightsone(double * response,
                                                 double * linpred)
    {
        if (counter==0)
        {
            set_worklin();
        }

        double l;
        double x = (*response);
        double xi = (*linpred);
        double omega = exp(*worklin[2]);
        double alpha = *worklin[1];
        double df = exp(*worklin[0]);
        double z = (x - xi) / omega;
        double tau = sqrt((df+1)/(pow(z,2) + df));

        l = - 0.5*(df+1) * log(1 + pow(z,2)/df) + log(gsl_cdf_tdist_P(alpha*z*tau, max(30.0, df+1)));

        modify_worklin();

        return l;

    }


    void DISTR_st_xi::compute_iwls_wweightschange_weightsone(
                                                             double * response,
                                                             double * linpred,
                                                             double * workingweight,
                                                             double * workingresponse,
                                                             double & like,
                                                             const bool & compute_like)
    {
        if (counter==0)
        {
            set_worklin();
        }
        double x = (*response);
        double xi = (*linpred);
        double omega = exp(*worklin[2]);
        double alpha = *worklin[1];
        double df = exp(*worklin[0]);
        double z = (x - xi) / omega;
        double tau = sqrt((df+1)/(pow(z,2) + df));

        double nu = z * pow(tau,2) / omega - alpha * tau * df / (pow(z,2) + df) / omega * gsl_ran_tdist_pdf(alpha*z*tau, max(30.0,df + 1)) / gsl_cdf_tdist_P(alpha*z*tau, max(30.0, df + 1)); // v

        *workingweight = ef[(max(-100, min(100, (int)nearbyint((*worklin[1])*10.0))) + 100)][(max(-20, min(50, (int)nearbyint((*worklin[0])*10.0))) + 20)] / pow(omega, 2); // w

        *workingresponse = *linpred + nu/(*workingweight);

        if (compute_like)
        {

            like += - 0.5*(df+1) * log(1 + pow(z,2)/df) + log(gsl_cdf_tdist_P(alpha*z*tau, max(30.0, df+1))); // loglikelihood

        }


        modify_worklin();

    }


    // ONLY FOR THE MAIN PARAMETER:
    void DISTR_st_xi::compute_mu_mult(vector<double *> linpred,vector<double *> response,double * mu)
    {
        *mu = ((*linpred[predstart_mumult+3])); // <--- depends on the number of parameters
    }


    void DISTR_st_xi::outoptions(void)
    {
        DISTR::outoptions();
        optionsp->out("Link function (xi): identity\n");
        optionsp->out("\n");
        optionsp->out("\n");
    }


    void DISTR_st_xi::update_end(void)
    {


        // helpmat1 stores (eta_mu)

        double * worklin;
        if (linpred_current==1)
            worklin = linearpred1.getV();
        else
            worklin = linearpred2.getV();

        double * pmu = helpmat1.getV();

        unsigned i;
        for (i=0;i<nrobs;i++,pmu++,worklin++)
        {
            *pmu = (*worklin);
            //    double t = 0;
        }

    }



        //------------------------------------------------------------------------------
        //-------------- CLASS: DISTR_bivst_df --------------------------------------
        //------------------------------------------------------------------------------

        // Constructor:
        DISTR_bivst_df::DISTR_bivst_df(GENERAL_OPTIONS * o,
                                             const datamatrix & r,
                                             const datamatrix & w)
        : DISTR_gamlss(o,r,7,w) // <-----
        {
            family = "Bivariate Skew-t Distribution - DF";
            outpredictor = true;
            outexpectation = false; // only true if main parameter
            predictor_name = "df";

            STOREDVALUES::get_bivst_ef_df_values(ef);
            STOREDVALUES::get_bivst_iota_values(iota);

            linpredminlimit=-3; // necessary for log-link
            linpredmaxlimit=6; // necessary for log-link

            // updateIWLS = false; // only false and necessary if gibbs sampling?
        }


        // Copy Constructor:
        DISTR_bivst_df::DISTR_bivst_df(const DISTR_bivst_df & nd)
        : DISTR_gamlss(DISTR_gamlss(nd))
        {
            response2 = nd.response2;
            response2p = nd.response2p;
            STOREDVALUES::get_bivst_ef_df_values(ef);
            STOREDVALUES::get_bivst_iota_values(iota);
        }

        // Operator:
        const DISTR_bivst_df & DISTR_bivst_df::operator=(
                                                               const DISTR_bivst_df & nd)
        {
            if (this==&nd)
                return *this;
            DISTR_gamlss::operator=(DISTR_gamlss(nd));
            response2 = nd.response2;
            response2p = nd.response2p;
            STOREDVALUES::get_bivst_ef_df_values(ef);
            STOREDVALUES::get_bivst_iota_values(iota);
            return *this;
        }


        double DISTR_bivst_df::get_intercept_start(void)
        {
            return 0;
        }

        void DISTR_bivst_df::compute_param_mult(vector<double *>  linpred,double * param)
        {
            *param = exp(*linpred[0]); // <--- c.f. linpred info
        }


        void DISTR_bivst_df::set_worklin(void)
        {

            DISTR_gamlss::set_worklin();

            response2p = response2.getV();

        }



        void DISTR_bivst_df::modify_worklin(void)
        {

            DISTR_gamlss::modify_worklin();

            if (counter<nrobs-1)
            {
                response2p++;
            }

        }


        double DISTR_bivst_df::loglikelihood_weightsone(double * response,
                                                           double * linpred)
        {
            if (counter==0)
            {
                set_worklin();
            }

            double df = exp(*linpred);
            double al2 = (*worklin[0]);
            double al1 = (*worklin[1]);
            double r = (*worklin[2]) / sqrt(1 + pow((*worklin[2]), 2));
            double r2 = r * r;
            double om2 = exp((*worklin[3]));
            double om1 = exp((*worklin[4]));
            double xi2 = (*worklin[5]);
            double xi1 = (*worklin[6]);

            double oneminrhotwo = 1.0 - r2;
            double z1 = (*response - xi1) / om1;
            double z2 = (*response2p - xi2) / om2;
            double Q = (z1 * z1 - 2 * r * z1 * z2 + z2 * z2) / oneminrhotwo;
            double dfplustwohalf = (df + 2.0) * 0.5;
            double dfhalf = df * 0.5;

            double l = 0;

            l += - log(df);
            l += randnumbers::lngamma_exact(dfplustwohalf) - randnumbers::lngamma_exact(dfhalf) - dfplustwohalf * log(1.0 + Q / df);
            l += log(gsl_cdf_tdist_P((al1 * z1 + al2 * z2) * sqrt((df + 2.0) / (df + Q)), max(30.0, df + 2.0))); // ll, simplified

            modify_worklin();

            return l;

        }


        void DISTR_bivst_df::compute_iwls_wweightschange_weightsone(
                                                                       double * response,
                                                                       double * linpred,
                                                                       double * workingweight,
                                                                       double * workingresponse,
                                                                       double & like,
                                                                       const bool & compute_like)
        {
            if (counter==0)
            {
                set_worklin();
               // cout << *linpred << endl;
            }

            *linpred = min(7.0, max(-4.0, *linpred));

            double df = exp(*linpred);
            double al2 = (*worklin[0]);
            double al1 = (*worklin[1]);
            double r = (*worklin[2]) / sqrt(1 + pow((*worklin[2]), 2));
            double om2 = exp((*worklin[3]));
            double om1 = exp((*worklin[4]));
            double xi2 = (*worklin[5]);
            double xi1 = (*worklin[6]);

            double r2 = r * r;

            double oneminrhotwo = 1.0 - r2;
            double z1 = (*response - xi1) / om1;
            double z2 = (*response2p - xi2) / om2;
            double Q = (z1 * z1 - 2 * r * z1 * z2 + z2 * z2) / oneminrhotwo;
            double dfplustwohalf = (df + 2.0) * 0.5;
            double dfhalf = df * 0.5;

            double tmp = (al1 * z1 + al2 * z2) * sqrt((df + 2.0) / (df + Q));

            double tval = gsl_ran_tdist_pdf(tmp, max(30.0,df + 2.0));
            double Tval = gsl_cdf_tdist_P(tmp, max(30.0, df + 2.0));

            double io = iota[max(-500, min(500, (int)nearbyint(tmp*10.0))) + 500][max(-20, min(50, (int)nearbyint((*linpred) * 10))) + 20];

            double li = - log(df) + randnumbers::lngamma_exact(dfplustwohalf) - randnumbers::lngamma_exact(dfhalf) - dfplustwohalf * log(1.0 + Q / df) + log(Tval);

            double dfp = exp(*linpred + 1e-5);
            double dfpplustwohalf = (dfp + 2) * 0.5;
            double dfphalf = dfp * 0.5;
            double Tvalp = gsl_cdf_tdist_P(tmp, max(30.0, dfp + 2.0));

            double lip = - log(dfp)+ randnumbers::lngamma_exact(dfpplustwohalf) - randnumbers::lngamma_exact(dfphalf) - dfpplustwohalf * log(1.0 + Q / dfp) + log(Tvalp);

//            double nu = (- 1 / df + 0.5 * (randnumbers::digamma_exact(dfplustwohalf)
//                                           - randnumbers::digamma_exact(dfhalf)
//                                           - log(1 + Q / df)
//                                           + (df + 2.0) * Q / (df + Q) / df
//                                           + tval / Tval * ((al1 * z1 + al2 * z2) * (1.0 / sqrt((df + 2) / (df + Q)) - sqrt(df + 2) / sqrt(df + Q) / (df + Q))) + io / Tval)) * df; // v


            double nu = (lip - li) / 1e-5;



            if(nu != nu){
                nu = 0.0;
            } // prevent nans (might only happen if *workingweight is fixed to a small value

            *workingweight = ef[max(-200, min(500, (int)nearbyint((*linpred) * 100))) + 200] * pow(df, 2); // w

            *workingresponse = *linpred + nu/(*workingweight);

            if (compute_like)
            {

                like += li;//- log(df) + randnumbers::lngamma_exact(dfplustwohalf) - randnumbers::lngamma_exact(dfhalf) - dfplustwohalf * log(1.0 + Q / df) + log(Tval); // loglikelihood

            }

            modify_worklin();

        }

        void DISTR_bivst_df::outoptions(void)
        {
            DISTR::outoptions();
            optionsp->out("  Link function (df): log\n");
            optionsp->out("\n");
            optionsp->out("\n");
        }



        void DISTR_bivst_df::update_end(void)
        {


            // helpmat1 stores (eta_mu)

            double * worklin;
            if (linpred_current==1)
                worklin = linearpred1.getV();
            else
                worklin = linearpred2.getV();

            double * pmu = helpmat1.getV();

            unsigned i;
            for (i=0;i<nrobs;i++,pmu++,worklin++)
            {
                *pmu = exp(*worklin);
                //    double t = 0;
            }

        }



        //------------------------------------------------------------------------------
        //-------------- CLASS: DISTR_bivst_alpha --------------------------------------
        //------------------------------------------------------------------------------

        // Constructor:
        DISTR_bivst_alpha::DISTR_bivst_alpha(GENERAL_OPTIONS * o,
                                             const datamatrix & r,
                                             unsigned & p,
                                             const datamatrix & w)
        : DISTR_gamlss(o,r,7,w) // <-----
        {
            pos = p;
            family = "Bivariate Skew-t Distribution - Alpha";
            outpredictor = true;
            outexpectation = false; // only true if main parameter
            predictor_name = "alpha";

            STOREDVALUES::get_bivst_ef_alpha_values(ef);

            linpredminlimit=-25; // necessary for log-link
            linpredmaxlimit=25; // necessary for log-link

            // updateIWLS = false; // only false and necessary if gibbs sampling?
        }


        // Copy Constructor:
        DISTR_bivst_alpha::DISTR_bivst_alpha(const DISTR_bivst_alpha & nd)
        : DISTR_gamlss(DISTR_gamlss(nd))
        {
            pos = nd.pos;
            response2 = nd.response2;
            response2p = nd.response2p;
            STOREDVALUES::get_bivst_ef_alpha_values(ef);
        }

        // Operator:
        const DISTR_bivst_alpha & DISTR_bivst_alpha::operator=(
                                                               const DISTR_bivst_alpha & nd)
        {
            if (this==&nd)
                return *this;
            DISTR_gamlss::operator=(DISTR_gamlss(nd));
            pos = nd.pos;
            response2 = nd.response2;
            response2p = nd.response2p;
            STOREDVALUES::get_bivst_ef_alpha_values(ef);
            return *this;
        }


        double DISTR_bivst_alpha::get_intercept_start(void)
        {
            return 0;
        }

        void DISTR_bivst_alpha::compute_param_mult(vector<double *>  linpred,double * param)
        {
            *param = (*linpred[1 + pos]); // <--- c.f. linpred info
        }


        void DISTR_bivst_alpha::set_worklin(void)
        {

            DISTR_gamlss::set_worklin();

            response2p = response2.getV();

        }



        void DISTR_bivst_alpha::modify_worklin(void)
        {

            DISTR_gamlss::modify_worklin();

            if (counter<nrobs-1)
            {
                response2p++;
            }

        }


        double DISTR_bivst_alpha::loglikelihood_weightsone(double * response,
                                                           double * linpred)
        {
            if (counter==0)
            {
                set_worklin();
            }

            double df = exp(*worklin[0]);
            double al2 = (*worklin[1]);
            double al1 = (*linpred);
            double r = (*worklin[2]) / sqrt(1 + pow((*worklin[2]), 2));
            double r2 = r * r;
            double om2 = exp((*worklin[3]));
            double om1 = exp((*worklin[4]));
            double xi2 = (*worklin[5]);
            double xi1 = (*worklin[6]);

            double oneminrhotwo = 1.0 - r2;
            double z1 = (*response - xi1) / om1;
            double z2 = (*response2p - xi2) / om2;
            double Q = (z1 * z1 - 2 * r * z1 * z2 + z2 * z2) / oneminrhotwo;
            double dfplustwohalf = (df + 2.0) * 0.5;

            double l = 0.0;

            l += log(gsl_cdf_tdist_P((al1 * z1 + al2 * z2) * sqrt((df + 2.0) / (df + Q)), max(30.0, df + 2.0)));
            // ll, simplified

            modify_worklin();

            return l;

        }


        void DISTR_bivst_alpha::compute_iwls_wweightschange_weightsone(
                                                                       double * response,
                                                                       double * linpred,
                                                                       double * workingweight,
                                                                       double * workingresponse,
                                                                       double & like,
                                                                       const bool & compute_like)
        {
            if (counter==0)
            {
                set_worklin();
            }

            double df = exp(*worklin[0]);
            double al2 = (*worklin[1]);
            double al1 = (*linpred);
            double r = (*worklin[2]) / sqrt(1 + pow((*worklin[2]), 2));
            double om2 = exp((*worklin[3]));
            double om1 = exp((*worklin[4]));
            double xi2 = (*worklin[5]);
            double xi1 = (*worklin[6]);

            double r2 = r * r;

            double oneminrhotwo = 1.0 - r2;
            double z1 = (*response - xi1) / om1;
            double z2 = (*response2p - xi2) / om2;
            double Q = (z1 * z1 - 2 * r * z1 * z2 + z2 * z2) / oneminrhotwo;
            double dfplustwohalf = (df + 2.0) * 0.5;

            double tval = gsl_ran_tdist_pdf((al1 * z1 + al2 * z2) * sqrt((df + 2.0) / (df + Q)), max(30.0, df + 2.0));
            double Tval = gsl_cdf_tdist_P((al1 * z1 + al2 * z2) * sqrt((df + 2.0) / (df + Q)), max(30.0, df + 2.0));

            double nu = tval / Tval * z1 * sqrt((df + 2.0) / (df + Q)); // v

            if(nu != nu){
                nu = 0.0;
            } // prevent nans (might only happen if *workingweight is fixed to a small value

            *workingweight = ef[max(-9, min(9, (int)nearbyint(r * 10))) + 9][max(-5, min(5, (int)nearbyint(al1*0.5))) + 5][max(-5, min(5, (int)nearbyint(al2*0.5))) + 5][max(-10, min(24, (int)nearbyint(log(df) * 5))) + 10]; // w

            *workingresponse = *linpred + nu/(*workingweight);

            if (compute_like)
            {

                like +=  log(Tval); // loglikelihood

            }

            modify_worklin();

        }

        void DISTR_bivst_alpha::outoptions(void)
        {
            DISTR::outoptions();
            optionsp->out("  Link function (alpha): identity\n");
            optionsp->out("\n");
            optionsp->out("\n");
        }



        void DISTR_bivst_alpha::update_end(void)
        {


            // helpmat1 stores (eta_mu)

            double * worklin;
            if (linpred_current==1)
                worklin = linearpred1.getV();
            else
                worklin = linearpred2.getV();

            double * pmu = helpmat1.getV();

            unsigned i;
            for (i=0;i<nrobs;i++,pmu++,worklin++)
            {
                *pmu = (*worklin);
                //    double t = 0;
            }

        }




        //------------------------------------------------------------------------------
        //-------------- CLASS: DISTR_bivst_rho ----------------------------------------
        //------------------------------------------------------------------------------

        // Constructor:
        DISTR_bivst_rho::DISTR_bivst_rho(GENERAL_OPTIONS * o,
                                         const datamatrix & r,
                                         const datamatrix & w)
        : DISTR_gamlss(o,r,7,w) // <-----
        {
            family = "Bivariate Skew-t Distribution - Rho";
            outpredictor = true;
            outexpectation = false; // only true if main parameter
            predictor_name = "rho";

            STOREDVALUES::get_bivst_ef_rho_values(ef);

            linpredminlimit=-100; // necessary for log-link
            linpredmaxlimit=100; // necessary for log-link

            // updateIWLS = false; // only false and necessary if gibbs sampling?
        }


        // Copy Constructor:
        DISTR_bivst_rho::DISTR_bivst_rho(const DISTR_bivst_rho & nd)
        : DISTR_gamlss(DISTR_gamlss(nd))
        {
            response2 = nd.response2;
            response2p = nd.response2p;
            STOREDVALUES::get_bivst_ef_rho_values(ef);
        }

        // Operator:
        const DISTR_bivst_rho & DISTR_bivst_rho::operator=(
                                                           const DISTR_bivst_rho & nd)
        {
            if (this==&nd)
                return *this;
            DISTR_gamlss::operator=(DISTR_gamlss(nd));
            response2 = nd.response2;
            response2p = nd.response2p;
            STOREDVALUES::get_bivst_ef_rho_values(ef);
            return *this;
        }

        double DISTR_bivst_rho::get_intercept_start(void)
        {
            return 0;
        }

        void DISTR_bivst_rho::compute_param_mult(vector<double *>  linpred,double * param)
        {
            double arg = (*linpred[3]);
            *param = arg/pow(1+pow(arg,2),0.5); // <--- c.f. linpred info
        }


        void DISTR_bivst_rho::set_worklin(void)
        {

            DISTR_gamlss::set_worklin();

            response2p = response2.getV();

        }



        void DISTR_bivst_rho::modify_worklin(void)
        {

            DISTR_gamlss::modify_worklin();

            if (counter<nrobs-1)
            {
                response2p++;
            }

        }

        double DISTR_bivst_rho::loglikelihood_weightsone(double * response,
                                                         double * linpred)
        {
            if (counter==0)
            {
                set_worklin();
            }

            double r = 0;

            if (*linpred <= -100)
                r  = -0.99995;
            else if (*linpred >= 100)
                r  = 0.99995;
            else
                r = (*linpred)/pow((1+pow((*linpred),2)),0.5);

            double r2 = r * r;
            double df = exp(*worklin[0]);
            double al2 = (*worklin[1]);
            double al1 = (*worklin[2]);
            double om2 = exp((*worklin[3]));
            double om1 = exp((*worklin[4]));
            double xi2 = (*worklin[5]);
            double xi1 = (*worklin[6]);

            double oneminrhotwo = 1.0 - r2;
            double z1 = (*response - xi1) / om1;
            double z2 = (*response2p - xi2) / om2;
            double Q = (z1 * z1 - 2 * r * z1 * z2 + z2 * z2) / oneminrhotwo;
            double dfplustwohalf = (df + 2.0) * 0.5;

            double l = 0.0;

            l += - 0.5 * log(oneminrhotwo);
            l += - dfplustwohalf * log(1.0 + Q / df);
            l += log(gsl_cdf_tdist_P((al1 * z1 + al2 * z2) * sqrt((df + 2.0) / (df + Q)), max(30.0, df + 2.0)));
            // calculate log-likelihood (possibly simplified)

            modify_worklin();

            return l;

        }


        void DISTR_bivst_rho::compute_iwls_wweightschange_weightsone(
                                                                     double * response,
                                                                     double * linpred,
                                                                     double * workingweight,
                                                                     double * workingresponse,
                                                                     double & like,
                                                                     const bool & compute_like)
        {

            if (counter==0)
            {
                set_worklin();
            }

            double r = 0;

            if (*linpred <= -100)
                r  = -0.99995;
            else if (*linpred >= 100)
                r  = 0.99995;
            else
                r = (*linpred)/pow((1+pow((*linpred),2)),0.5);

            double r2 = r * r;
            double df = exp(*worklin[0]);
            double al2 = (*worklin[1]);
            double al1 = (*worklin[2]);
            double om2 = exp((*worklin[3]));
            double om1 = exp((*worklin[4]));
            double xi2 = (*worklin[5]);
            double xi1 = (*worklin[6]);

            double oneminrhotwo = 1.0 - r2;
            double z1 = (*response - xi1) / om1;
            double z2 = (*response2p - xi2) / om2;
            double Q = (z1 * z1 - 2 * r * z1 * z2 + z2 * z2) / oneminrhotwo;
            double dfplustwohalf = (df + 2.0) * 0.5;

            double dlinkf = 1.0 / sqrt(1.0 + pow(max(-100.0, min(100.0, (*linpred))), 2)) * (1 - r2);

            double tval = gsl_ran_tdist_pdf((al1 * z1 + al2 * z2) * sqrt((df + 2.0) / (df + Q)), max(30.0,df + 2.0));
            double Tval = gsl_cdf_tdist_P((al1 * z1 + al2 * z2) * sqrt((df + 2.0) / (df + Q)), max(30.0, df + 2.0));

            double nu = (r / oneminrhotwo + (- (df + 2.0) / (df + Q) + tval / Tval * (al1 * z1 + al2 * z2) * sqrt((df + 2.0) / (df + Q)) / (df + Q)) * (r / oneminrhotwo / oneminrhotwo * (z1 * z1 - 2.0 * r * z1 * z2 + z2 * z2) - 1 / oneminrhotwo * z1 * z2)) * dlinkf; // v

            if(nu != nu){
                nu = 0.0;
            } // prevent nans (might only happen if *workingweight is fixed to a small value

            *workingweight = ef[max(-9, min(9, (int)nearbyint(r * 10))) + 9][max(-5, min(5, (int)nearbyint(al1*0.5))) + 5][max(-5, min(5, (int)nearbyint(al2*0.5))) + 5][max(-10, min(24, (int)nearbyint(log(df) * 5))) + 10] * dlinkf * dlinkf; // w

            *workingresponse = *linpred + nu/(*workingweight);

            if (compute_like)
            {

                like += - 0.5 * log(oneminrhotwo) - dfplustwohalf * log(1.0 + Q / df) + log(Tval);

            }


            modify_worklin();

        }

        void DISTR_bivst_rho::outoptions(void)
        {
            DISTR::outoptions();
            optionsp->out("  Response function (rho): rhogit\n");
            optionsp->out("\n");
            optionsp->out("\n");
        }


        void DISTR_bivst_rho::update_end(void)
        {


            // helpmat1 stores (eta_mu)

            double * worklin;
            if (linpred_current==1)
                worklin = linearpred1.getV();
            else
                worklin = linearpred2.getV();

            double * pmu = helpmat1.getV();

            unsigned i;
            for (i=0;i<nrobs;i++,pmu++,worklin++)
            {
                *pmu = (*worklin)/pow((1+pow((*worklin),2)),0.5);
                //    double t = 0;
            }

        }


        //------------------------------------------------------------------------------
        //-------------- CLASS: DISTR_bivst_omega --------------------------------------
        //------------------------------------------------------------------------------

        // Constructor:
        DISTR_bivst_omega::DISTR_bivst_omega(GENERAL_OPTIONS * o,
                                             const datamatrix & r,
                                             unsigned &p,
                                             const datamatrix & w)
        : DISTR_gamlss(o,r,7,w) // <-----
        {
            pos = p;
            family = "Bivariate Skew-t Distribution - Omega";
            outpredictor = true;
            outexpectation = false; // only true if main parameter
            predictor_name = "omega";

            STOREDVALUES::get_bivst_ef_omega_values(ef);

            linpredminlimit=-10; // necessary for log-link
            linpredmaxlimit=15; // necessary for log-link

            // updateIWLS = false; // only false and necessary if gibbs sampling?
        }


        // Copy Constructor:
        DISTR_bivst_omega::DISTR_bivst_omega(const DISTR_bivst_omega & nd)
        : DISTR_gamlss(DISTR_gamlss(nd))
        {
            pos = nd.pos;
            response2 = nd.response2;
            response2p = nd.response2p;
            STOREDVALUES::get_bivst_ef_omega_values(ef);
        }

        // Operator:
        const DISTR_bivst_omega & DISTR_bivst_omega::operator=(
                                                               const DISTR_bivst_omega & nd)
        {
            if (this==&nd)
                return *this;
            DISTR_gamlss::operator=(DISTR_gamlss(nd));
            pos = nd.pos;
            response2 = nd.response2;
            response2p = nd.response2p;
            STOREDVALUES::get_bivst_ef_omega_values(ef);
            return *this;
        }

        double DISTR_bivst_omega::get_intercept_start(void)
        {
            return 0;
        }

        void DISTR_bivst_omega::compute_param_mult(vector<double *>  linpred,double * param)
        {
            *param = exp(*linpred[4 + pos]); // <--- c.f. linpred info
        }

        void DISTR_bivst_omega::set_worklin(void)
        {

            DISTR_gamlss::set_worklin();

            response2p = response2.getV();

        }



        void DISTR_bivst_omega::modify_worklin(void)
        {

            DISTR_gamlss::modify_worklin();

            if (counter<nrobs-1)
            {
                response2p++;
            }

        }

        double DISTR_bivst_omega::loglikelihood_weightsone(double * response,
                                                           double * linpred)
        {
            if (counter==0)
            {
                set_worklin();
            }

            double df = exp(*worklin[0]);
            double al2 = (*worklin[1]);
            double al1 = (*worklin[2]);
            double r = (*worklin[3]) / sqrt(1 + pow((*worklin[3]), 2));
            double r2 = r * r;
            double om2 = exp((*worklin[4]));
            double om1 = exp((*linpred));
            double xi2 = (*worklin[5]);
            double xi1 = (*worklin[6]);

            double oneminrhotwo = 1.0 - r2;
            double z1 = (*response - xi1) / om1;
            double z2 = (*response2p - xi2) / om2;
            double Q = (z1 * z1 - 2 * r * z1 * z2 + z2 * z2) / oneminrhotwo;
            double dfplustwohalf = (df + 2.0) * 0.5;

            double l;

            l += - log(om1);
            l += - dfplustwohalf * log(1.0 + Q / df);
            l += log(gsl_cdf_tdist_P((al1 * z1 + al2 * z2) * sqrt((df + 2.0) / (df + Q)), max(30.0, df + 2.0)));
            // calculate log-likelihood (possibly simplified)

            modify_worklin();

            return l;

        }


        void DISTR_bivst_omega::compute_iwls_wweightschange_weightsone(
                                                                       double * response,
                                                                       double * linpred,
                                                                       double * workingweight,
                                                                       double * workingresponse,
                                                                       double & like,
                                                                       const bool & compute_like)
        {
            if (counter==0)
            {
                set_worklin();
            }

            double df = exp(*worklin[0]);
            double al2 = (*worklin[1]);
            double al1 = (*worklin[2]);
            double r = (*worklin[3]) / sqrt(1 + pow((*worklin[3]), 2));
            double r2 = r * r;
            double om2 = exp((*worklin[4]));
            double om1 = exp((*linpred));
            double xi2 = (*worklin[5]);
            double xi1 = (*worklin[6]);

            double oneminrhotwo = 1.0 - r2;
            double z1 = (*response - xi1) / om1;
            double z2 = (*response2p - xi2) / om2;
            double Q = (z1 * z1 - 2 * r * z1 * z2 + z2 * z2) / oneminrhotwo;
            double dfplustwohalf = (df + 2.0) * 0.5;

            double tval = gsl_ran_tdist_pdf((al1 * z1 + al2 * z2) * sqrt((df + 2.0) / (df + Q)), max(30.0, df + 2.0));
            double Tval = gsl_cdf_tdist_P((al1 * z1 + al2 * z2) * sqrt((df + 2.0) / (df + Q)), max(30.0, df + 2.0));

            double nu = (- 1.0 - z1 * (- (df + 2.0) / 2.0 * (2.0 * z1 - 2 * r * z2) / (1.0 + Q / df) / df / oneminrhotwo
                         + tval / Tval * (al1 * sqrt((df + 2.0) / (df + Q)) - (al1 * z1 + al2 * z2) * sqrt(df + 2.0) * (z1 - r * z2) / sqrt(df + Q) / (df + Q)))); // v

            if(nu != nu){
                nu = 0.0;
            } // prevent nans (might only happen if *workingweight is fixed to a small value

            *workingweight = ef[max(-9, min(9, (int)nearbyint(r * 10))) + 9][max(-5, min(5, (int)nearbyint(al1*0.5))) + 5][max(-5, min(5, (int)nearbyint(al2*0.5))) + 5][max(-10, min(24, (int)nearbyint(log(df) * 5))) + 10]; // w

            *workingresponse = *linpred + nu/(*workingweight);

            if (compute_like)
            {

                like += - log(om1) - dfplustwohalf * log(1.0 + Q / df) + log(Tval); // loglikelihood

            }


            modify_worklin();

        }

        void DISTR_bivst_omega::outoptions(void)
        {
            DISTR::outoptions();
            optionsp->out("  Link function (omega): log\n");
            optionsp->out("\n");
            optionsp->out("\n");
        }


        void DISTR_bivst_omega::update_end(void)
        {


            // helpmat1 stores (eta_mu)

            double * worklin;
            if (linpred_current==1)
                worklin = linearpred1.getV();
            else
                worklin = linearpred2.getV();

            double * pmu = helpmat1.getV();

            unsigned i;
            for (i=0;i<nrobs;i++,pmu++,worklin++)
            {
                *pmu = exp(*worklin);
                //    double t = 0;
            }

        }


        //------------------------------------------------------------------------------
        //-------------- CLASS: DISTR_bivst_xi -----------------------------------------
        //------------------------------------------------------------------------------

        // ONLY FOR THE MAIN PARAMETER:
        void DISTR_bivst_xi::check_errors(void)
        {

            if (errors==false)
            {
                unsigned i=0;
                double * workresp = response.getV();
                double * workweight = weight.getV();
                while ( (i<nrobs) && (errors==false) )
                {

                    if (*workweight > 0)
                    {


                    }
                    else if (*workweight == 0)
                    {
                    }
                    else
                    {
                        errors=true;
                        errormessages.push_back("ERROR: negative weights encountered\n");
                    }

                    i++;
                    workresp++;
                    workweight++;

                }

            }

        }

        // Constructor:
        DISTR_bivst_xi::DISTR_bivst_xi(GENERAL_OPTIONS * o,
                                       const datamatrix & r,
                                       unsigned &p,
                                       const datamatrix & w)
        : DISTR_gamlss(o,r,7,w) // <-----
        {
            pos = p;
            family = "Bivariate Skew-t Distribution - Xi";
            outpredictor = true;
            outexpectation = true; // only true if main parameter
            predictor_name = "xi";

            STOREDVALUES::get_bivst_ef_xi_values(ef);

            // linpredminlimit=-10; // necessary for log-link
            // linpredmaxlimit=15; // necessary for log-link

            // updateIWLS = false; // only false and necessary if gibbs sampling?
        }


        // Copy Constructor:
        DISTR_bivst_xi::DISTR_bivst_xi(const DISTR_bivst_xi & nd)
        : DISTR_gamlss(DISTR_gamlss(nd))
        {
            pos = nd.pos;
            response2 = nd.response2;
            response2p = nd.response2p;
            STOREDVALUES::get_bivst_ef_xi_values(ef);
        }

        // Operator:
        const DISTR_bivst_xi & DISTR_bivst_xi::operator=(
                                                         const DISTR_bivst_xi & nd)
        {
            if (this==&nd)
                return *this;
            DISTR_gamlss::operator=(DISTR_gamlss(nd));
            pos = nd.pos;
            response2 = nd.response2;
            response2p = nd.response2p;
            STOREDVALUES::get_bivst_ef_xi_values(ef);
            return *this;
        }


        // ONLY FOR THE MAIN PARAMETER:
        void DISTR_bivst_xi::compute_deviance_mult(vector<double *> response,
                                                   vector<double *> weight,
                                                   vector<double *> linpred,
                                                   double * deviance,
                                                   vector<datamatrix*> aux)
        {
            if (*weight[7] == 0) // <---
                *deviance=0;
            else
            {
                double df = exp(*linpred[0]);
                double al2 = (*linpred[1]);
                double al1 = (*linpred[2]);
                double r = (*linpred[3]) / sqrt(1 + pow((*linpred[3]), 2));
                double r2 = r * r;
                double om2 = exp((*linpred[4]));
                double om1 = exp((*linpred[5]));
                double xi2 = (*linpred[6]);
                double xi1 = (*linpred[7]);

                double oneminrhotwo = 1.0 - r2;
                double z1 = (*response[7] - xi1) / om1;
                double z2 = (*response[6] - xi2) / om2;
                double Q = (z1 * z1 - 2 * r * z1 * z2 + z2 * z2) / oneminrhotwo;
                double dfplustwohalf = (df + 2.0) * 0.5;
                double dfhalf = df * 0.5;

                double l = 0;

                l += log(2.0) - log(df) - log(PI) - log(om1) - log(om2) - 0.5 * log(oneminrhotwo);
                l += randnumbers::lngamma_exact(dfplustwohalf) - randnumbers::lngamma_exact(dfhalf) - dfplustwohalf * log(1.0 + Q / df);
                l += log(gsl_cdf_tdist_P((al1 * z1 + al2 * z2) * sqrt((df + 2.0) / (df + Q)), max(30.0, df + 2.0)));
                // Compute log-likelihood (complete)

                *deviance = -2*l;
            }
        }


        double DISTR_bivst_xi::get_intercept_start(void)
        {
            return 0;
        }

        void DISTR_bivst_xi::compute_param_mult(vector<double *>  linpred,double * param)
        {
            *param = (*linpred[6+pos]); // <--- c.f. linpred info
        }

        // ONLY FOR THE MAIN PARAMETER:
        double DISTR_bivst_xi::cdf_mult(vector<double *> response,
                                        vector<double *> param,
                                        vector<double *> weight,
                                        vector<datamatrix *> aux)


        {
            return 0; // calculate cdf for observation using param!
        }

        // ONLY FOR THE MAIN PARAMETER:
        double DISTR_bivst_xi::pdf_mult(vector<double *> response,
                                        vector<double *> param,
                                        vector<double *> weight,
                                        vector<datamatrix *> aux)
        {
            return 0; // usually set to be zero
        }

        void DISTR_bivst_xi::set_worklin(void)
        {

            DISTR_gamlss::set_worklin();

            response2p = response2.getV();

        }



        void DISTR_bivst_xi::modify_worklin(void)
        {

            DISTR_gamlss::modify_worklin();

            if (counter<nrobs-1)
            {
                response2p++;
            }

        }


        double DISTR_bivst_xi::loglikelihood_weightsone(double * response,
                                                        double * linpred)
        {
            if (counter==0)
            {
                set_worklin();
            }

            double df = exp(*worklin[0]);
            double al2 = (*worklin[1]);
            double al1 = (*worklin[2]);
            double r = (*worklin[3]) / sqrt(1 + pow((*worklin[3]), 2));
            double r2 = r * r;
            double om2 = exp((*worklin[4]));
            double om1 = exp((*worklin[5]));
            double xi2 = (*worklin[6]);
            double xi1 = (*linpred);

            double oneminrhotwo = 1.0 - r2;
            double z1 = (*response - xi1) / om1;
            double z2 = (*response2p - xi2) / om2;
            double Q = (z1 * z1 - 2 * r * z1 * z2 + z2 * z2) / oneminrhotwo;
            double dfplustwohalf = (df + 2.0) * 0.5;

            double l = 0;

            l += - dfplustwohalf * log(1.0 + Q / df);
            l += log(gsl_cdf_tdist_P((al1 * z1 + al2 * z2) * sqrt((df + 2.0) / (df + Q)), max(30.0, df + 2.0)));

            modify_worklin();

            return l;

        }


        void DISTR_bivst_xi::compute_iwls_wweightschange_weightsone(
                                                                    double * response,
                                                                    double * linpred,
                                                                    double * workingweight,
                                                                    double * workingresponse,
                                                                    double & like,
                                                                    const bool & compute_like)
        {
            if (counter==0)
            {
                set_worklin();
            }

            double df = exp(*worklin[0]);
            double al2 = (*worklin[1]);
            double al1 = (*worklin[2]);
            double r = (*worklin[3]) / sqrt(1 + pow((*worklin[3]), 2));
            double r2 = r * r;
            double om2 = exp((*worklin[4]));
            double om1 = exp((*worklin[5]));
            double xi2 = (*worklin[6]);
            double xi1 = (*linpred);

            double oneminrhotwo = 1.0 - r2;
            double z1 = (*response - xi1) / om1;
            double z2 = (*response2p - xi2) / om2;
            double Q = (z1 * z1 - 2 * r * z1 * z2 + z2 * z2) / oneminrhotwo;
            double dfplustwohalf = (df + 2.0) * 0.5;

            double tval = gsl_ran_tdist_pdf((al1 * z1 + al2 * z2) * sqrt((df + 2.0) / (df + Q)), max(30.0, df + 2.0));
            double Tval = gsl_cdf_tdist_P((al1 * z1 + al2 * z2) * sqrt((df + 2.0) / (df + Q)), max(30.0, df + 2.0));

            double nu = - (- (df + 2.0) / 2.0 * (2.0 * z1 - 2 * r * z2) / (1.0 + Q / df) / df / oneminrhotwo
                           + tval / Tval * (al1 * sqrt((df + 2.0) / (df + Q)) - (al1 * z1 + al2 * z2) * sqrt(df + 2.0) * (z1 - r * z2) / oneminrhotwo / sqrt(df + Q) / (df + Q)));

            if(nu != nu){
                nu = 0.0;
            } // prevent nans (might only happen if *workingweight is fixed to a small value

            *workingweight = ef[max(-9, min(9, (int)nearbyint(r * 10))) + 9][max(-5, min(5, (int)nearbyint(al1*0.5))) + 5][max(-5, min(5, (int)nearbyint(al2*0.5))) + 5][max(-10, min(24, (int)nearbyint(log(df) * 5))) + 10] / pow(om1, 2); // w

            *workingresponse = *linpred + nu/(*workingweight);

            if (compute_like)
            {

                like += - dfplustwohalf * log(1.0 + Q / df) + log(Tval); // loglikelihood

            }


            modify_worklin();

        }


        // ONLY FOR THE MAIN PARAMETER:
        void DISTR_bivst_xi::compute_mu_mult(vector<double *> linpred,vector<double *> response,double * mu)
        {
            *mu = ((*linpred[predstart_mumult+5+pos])); // <--- depends on the number of parameters
        }


        void DISTR_bivst_xi::outoptions(void)
        {
            DISTR::outoptions();
            optionsp->out("  Response function (xi): identity\n");
            optionsp->out("\n");
            optionsp->out("\n");
        }


        void DISTR_bivst_xi::update_end(void)
        {


            // helpmat1 stores (eta_mu)

            double * worklin;
            if (linpred_current==1)
                worklin = linearpred1.getV();
            else
                worklin = linearpred2.getV();

            double * pmu = helpmat1.getV();

            unsigned i;
            for (i=0;i<nrobs;i++,pmu++,worklin++)
            {
                *pmu = (*worklin);
                //    double t = 0;:
            }
        }



}
